from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...common.responsive import ResponsiveMixin
from ...common.spacing import SpacingMixin
from .. import alert
from . import forms, models

mixin_factory = settings.get_renderer(alert)


@plugin_pool.register_plugin
class AlertPlugin(
    mixin_factory("Alert"), AttributesMixin, ResponsiveMixin, SpacingMixin, CMSUIPlugin
):
    """
    Components > "Alerts" Plugin
    https://getbootstrap.com/docs/5.0/components/alerts/
    """

    name = _("Alert")
    module = _("Frontend")
    model = models.Alert
    form = forms.AlertForm
    change_form_template = "djangocms_frontend/admin/alert.html"
    allow_children = True

    fieldsets = [
        (
            None,
            {
                "fields": [
                    "alert_context",
                    "alert_dismissible",
                ]
            },
        ),
    ]



from django.conf import settings
from django.utils.translation import gettext_lazy as _

CAROUSEL_PAUSE_CHOICES = (
    ("hover", _("On hover")),
    ("false", _("Never")),
)

CAROUSEL_TEMPLATE_CHOICES = getattr(
    settings,
    "DJANGOCMS_FRONTEND_CAROUSEL_TEMPLATES",
    (("default", _("Default")),),
)

# this is used when no valua is passed in the template via
# {% with 1024 as width and 768 as height %}
CAROUSEL_DEFAULT_SIZE = getattr(
    settings,
    "DJANGOCMS_FRONTEND_CAROUSEL_DEFAULT_SIZE",
    [1024, 768],
)

CAROUSEL_ASPECT_RATIOS = (
    (1, 1),
    (3, 2),
    (4, 3),
    (21, 9),
    (18, 9),
) + tuple(getattr(settings, "DJANGOCMS_FRONTEND_CAROUSEL_ASPECT_RATIOS", tuple()))

CAROUSEL_ASPECT_RATIO_CHOICES = tuple(
    (f"{x}x{y}", f"{x}x{y}") for x, y in CAROUSEL_ASPECT_RATIOS
)

CAROUSEL_TRANSITION_CHOICES = (
    ("", _("Slide")),
    ("carousel-fade", _("Fade")),
)



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...common.background import BackgroundMixin
from ...common.responsive import ResponsiveMixin
from ...common.spacing import SpacingMixin
from .. import icon
from . import forms, models

mixin_factory = settings.get_renderer(icon)


class IconPlugin(
    mixin_factory("Icon"),
    AttributesMixin,
    ResponsiveMixin,
    SpacingMixin,
    BackgroundMixin,
    CMSUIPlugin,
):
    """
    Universal icon picker
    https://github.com/migliori/universal-icon-picker
    """

    name = _("Icon")
    module = _("Frontend")
    model = models.Icon
    form = forms.IconForm
    text_enabled = True

    fieldsets = [
        (
            None,
            {
                "fields": (
                    (
                        "icon",
                        "icon_size",
                    ),
                    "icon_foreground",
                    "icon_rounded",
                )
            },
        ),
    ]


if "IconPlugin" in plugin_pool.plugins:
    #  Unregister already installed IconPlugin
    del plugin_pool.plugins["IconPlugin"]

plugin_pool.register_plugin(IconPlugin)



"""
See PEP 440 (https://www.python.org/dev/peps/pep-0440/)

Release logic:
 1. Increase version number (change __version__ below).
 2. Ensure the static bundle is upto date with ``nvm use && npm install && gulp sass && gulp bundle``
 3. Assure that all changes have been documented in CHANGELOG.rst.
 4. In setup.py check that
   - versions from all third party packages are pinned in ``REQUIREMENTS``.
   - the list of ``CLASSIFIERS`` is up-to-date.
 5. git add djangocms_frontend/__init__.py CHANGELOG.rst setup.py
 6. run ``django-admin compilemessages`` from the ``djangocms_frontend`` folder
 7. git add all changed .mo files
 8. git commit -m 'Bump to {new version}'
 9. git push
10. Assure that all tests pass on https://github.com/django-cms/djangocms-frontend/actions
11. Create a new release on https://github.com/django-cms/djangocms-frontend/releases/new
12. Publish the release when ready
13. Github actions will publish the new package to pypi
"""

__version__ = "1.1.1"



from django.utils.translation import gettext_lazy as _

from djangocms_frontend.models import FrontendUIItem


class Jumbotron(FrontendUIItem):
    """
    Components > "Jumbotron" Plugin
    https://getbootstrap.com/docs/5.0/components/jumbotron/
    """

    class Meta:
        proxy = True
        verbose_name = _("Jumbotron")

    def get_short_description(self):
        text = ""
        if self.jumbotron_fluid:
            text = "({})".format(_("Fluid"))
        return text



from django import template

from djangocms_frontend import settings

from ..cms_plugins import create_tree

register = template.Library()


@register.inclusion_tag(
    f"djangocms_frontend/{settings.framework}/toc.html", takes_context=True
)
def table_of_contents(context):
    return dict(
        TOC=create_tree(context.get("request", []))
    )  # Empty TOC if no request object available



import os
from collections import OrderedDict

from django.core.management.base import BaseCommand, CommandParser
from django.core.management.color import color_style, no_style


def add_builtin_arguments(parser):
    parser.add_argument(
        "--noinput",
        action="store_false",
        dest="interactive",
        default=True,
        help="Tells Django CMS to NOT prompt the user for input of any kind.",
    )

    # These are taking "as-is" from Django's management base
    # management command.
    parser.add_argument(
        "-v",
        "--verbosity",
        action="store",
        dest="verbosity",
        default="1",
        type=int,
        choices=[0, 1, 2, 3],
        help="Verbosity level; 0=minimal output, 1=normal output, 2=verbose output, 3=very verbose output",
    )
    parser.add_argument(
        "--settings",
        help=(
            "The Python path to a settings module, e.g. "
            '"myproject.settings.main". If this isn\'t provided, the '
            "DJANGO_SETTINGS_MODULE environment variable will be used."
        ),
    )
    parser.add_argument(
        "--pythonpath",
        help='A directory to add to the Python path, e.g. "/home/djangoprojects/myproject".',
    )
    parser.add_argument(
        "--traceback", action="store_true", help="Raise on CommandError exceptions"
    )
    parser.add_argument(
        "--no-color",
        action="store_true",
        dest="no_color",
        default=False,
        help="Don't colorize the command output.",
    )
    parser.add_argument(
        "--force-color",
        action="store_true",
        dest="force_color",
        default=False,
        help="Colorize the command output.",
    )
    parser.add_argument(
        "--skip-checks",
        action="store_true",
        dest="skip_checks",
        default=False,
        help="Skip the checks.",
    )


class SubcommandsCommand(BaseCommand):
    subcommands = OrderedDict()
    instances = {}
    help_string = ""
    command_name = ""
    stealth_options = ("interactive",)

    subcommand_dest = "subcmd"

    def create_parser(self, prog_name, subcommand, **kwargs):
        kwargs = {}
        parser = CommandParser(
            prog=f"{os.path.basename(prog_name)} {subcommand}",
            description=self.help or None,
            **kwargs,
        )
        add_builtin_arguments(parser)
        self.add_arguments(parser)
        return parser

    def add_arguments(self, parser):
        self.instances = {}

        if self.subcommands:
            stealth_options = set(self.stealth_options)
            subparsers = parser.add_subparsers(dest=self.subcommand_dest)
            for command, cls in self.subcommands.items():
                instance = cls(self.stdout._out, self.stderr._out)
                instance.style = self.style
                kwargs = {}
                parser_sub = subparsers.add_parser(
                    name=instance.command_name,
                    help=instance.help_string,
                    description=instance.help_string,
                    **kwargs,
                )

                add_builtin_arguments(parser=parser_sub)
                instance.add_arguments(parser_sub)
                stealth_options.update({action.dest for action in parser_sub._actions})
                self.instances[command] = instance
            self.stealth_options = tuple(stealth_options)

    def handle(self, *args, **options):
        if options[self.subcommand_dest] in self.instances:
            command = self.instances[options[self.subcommand_dest]]
            if options.get("no_color"):
                command.style = no_style()
                command.stderr.style_func = None
            if options.get("force_color"):
                command.style = color_style(force_color=True)
            if options.get("stdout"):
                command.stdout._out = options.get("stdout")
            if options.get("stderr"):
                command.stderr._out = options.get("stderr")
            command.handle(*args, **options)
        else:
            self.print_help("manage.py", "frontend")



from django import forms
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelForm

from djangocms_frontend import settings
from djangocms_frontend.common.background import BackgroundFormMixin
from djangocms_frontend.contrib import navigation
from djangocms_frontend.contrib.link.forms import AbstractLinkForm, LinkForm
from djangocms_frontend.fields import (
    AttributesFormField,
    ButtonGroup,
    IconGroup,
    TemplateChoiceMixin,
)
from djangocms_frontend.helpers import first_choice
from djangocms_frontend.models import FrontendUIItem
from djangocms_frontend.settings import NAVBAR_DESIGNS

mixin_factory = settings.get_forms(navigation)


class NavigationForm(
    mixin_factory("Navigation"),
    BackgroundFormMixin,
    TemplateChoiceMixin,
    EntangledModelForm,
):
    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "template",
                "navbar_container",
                "navbar_design",
                "navbar_breakpoint",
                "attributes",
            ]
        }
        untangled_fields = ()

    template = forms.ChoiceField(
        label=_("Template"),
        choices=settings.NAVIGATION_TEMPLATE_CHOICES,
        initial=first_choice(settings.NAVIGATION_TEMPLATE_CHOICES),
        help_text=_("Defines the whole template set for this navigation."),
    )
    navbar_container = forms.BooleanField(
        label=_("Container"),
        required=False,
        initial=True,
    )
    navbar_design = forms.ChoiceField(
        label=_("Design"),
        required=True,
        choices=NAVBAR_DESIGNS,
        initial=first_choice(NAVBAR_DESIGNS),
        widget=ButtonGroup(attrs=dict(property="nav-design")),
    )
    navbar_breakpoint = forms.ChoiceField(
        label=_("Expand on device (and larger)"),
        required=False,
        choices=settings.EMPTY_CHOICE + settings.DEVICE_CHOICES,
        initial=settings.EMPTY_CHOICE[0][0],
        widget=IconGroup(),
    )
    attributes = AttributesFormField()


class PageTreeForm(mixin_factory("PageTree"), EntangledModelForm):
    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "start_level",
                "attributes",
            ]
        }
        untangled_fields = ()

    start_level = forms.IntegerField(
        label=_("Start level"),
        initial=0,
        help_text=_(
            "Start level of this page tree (0: root, 1: level below root, etc.)"
        ),
    )
    attributes = AttributesFormField()


class NavBrandForm(mixin_factory("NavBrand"), AbstractLinkForm, EntangledModelForm):
    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "simple_content",
                "attributes",
            ]
        }
        untangled_fields = ()

    link_is_optional = True

    simple_content = forms.CharField(
        label=_("Brand"),
        required=True,
        help_text=_("Enter brand name or add child plugins for brand icon or image"),
    )
    attributes = AttributesFormField()


class NavContainerForm(mixin_factory("NavContainer"), EntangledModelForm):
    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "attributes",
            ]
        }
        untangled_fields = ()

    attributes = AttributesFormField()


class NavLinkForm(mixin_factory("NavLink"), LinkForm):
    link_is_optional = True



class TabRenderMixin:
    def render(self, context, instance, placeholder):
        if instance.config.get("tab_alignment", "") == "flex-column":
            instance.add_classes("flex-row", "d-flex")
        return super().render(context, instance, placeholder)


class TabItemRenderMixin:
    def render(self, context, instance, placeholder):
        parent = instance.parent.get_plugin_instance()[0]
        instance.add_classes("tab-pane", parent.tab_effect)
        if instance.config.get("tab_bordered", False):
            if context["parent"].tab_type == "nav-tabs":
                instance.add_classes("border-end border-start border-bottom")
            else:
                instance.add_classes("border")
        if context["parent"].tab_type == "nav-tabs":
            instance.add_classes("bg-white")
        if parent.tab_index == context["parentloop"]["counter"]:
            instance.add_classes("show active")
        return super().render(context, instance, placeholder)



from cms.api import add_plugin
from cms.test_utils.testcases import CMSTestCase
from django.template.loader import select_template

from djangocms_frontend.contrib.carousel.cms_plugins import CarouselPlugin
from djangocms_frontend.contrib.carousel.constants import CAROUSEL_TEMPLATE_CHOICES
from djangocms_frontend.helpers import (
    first_choice,
    get_plugin_template,
    get_template_path,
    insert_fields,
    link_to_framework_doc,
)

from .fixtures import TestFixture


class HelpersTestCase(TestFixture, CMSTestCase):
    def test_get_template_path(self):
        template = get_template_path("carousel", "default", "slide")
        result = "djangocms_frontend/bootstrap5/carousel/default/slide.html"
        self.assertEqual(template, result)
        status = select_template([template])
        self.assertEqual(status.template.name, result)

    def test_get_plugin_template(self):
        self.superuser = self.get_superuser()
        self.language = "en"
        instance = add_plugin(
            placeholder=self.placeholder,
            plugin_type=CarouselPlugin.__name__,
            language="en",
        )
        template = get_plugin_template(
            instance,
            "carousel",
            "carousel",
            CAROUSEL_TEMPLATE_CHOICES,
        )
        self.assertEqual(
            template, "djangocms_frontend/bootstrap5/carousel/default/carousel.html"
        )
        # trigger default template
        template = get_plugin_template(
            instance,
            "does_not",
            "exist",
            CAROUSEL_TEMPLATE_CHOICES,
        )
        self.assertEqual(
            template, "djangocms_frontend/bootstrap5/does_not/default/exist.html"
        )
        # cleanup
        self.page.delete()

    def test_insert_fields(self):
        fieldsets = (
            (
                None,
                {
                    "fields": (
                        "F1",
                        "F2",
                        "F3",
                    )
                },
            ),
        )
        fs = insert_fields(fieldsets, ("F4",), block=0)
        self.assertEqual(
            fs,
            [
                (
                    None,
                    {
                        "fields": [
                            "F1",
                            "F2",
                            "F3",
                            "F4",
                        ],
                    },
                ),
            ],
        )
        fs = insert_fields(fieldsets, ("F4",), position=1, block=0)
        self.assertEqual(
            fs,
            [
                (
                    None,
                    {
                        "fields": [
                            "F1",
                            "F4",
                            "F2",
                            "F3",
                        ],
                    },
                ),
            ],
        )
        fs = insert_fields(fieldsets, ("F4",), position=-2, block=0)
        self.assertEqual(
            fs,
            [
                (
                    None,
                    {
                        "fields": [
                            "F1",
                            "F2",
                            "F4",
                            "F3",
                        ],
                    },
                ),
            ],
        )

        fs = insert_fields(fieldsets, ("F4",), blockname="Tarantula")
        self.assertEqual(
            fs,
            [
                (
                    None,
                    {
                        "fields": (
                            "F1",
                            "F2",
                            "F3",
                        ),
                    },
                ),
                (
                    "Tarantula",
                    {
                        "classes": ("collapse",),
                        "fields": [
                            "F4",
                        ],
                    },
                ),
            ],
        )

    def test_first_choice(self):
        self.assertEqual(
            first_choice(
                (
                    ("first", "First"),
                    ("second", "Second"),
                )
            ),
            "first",
        )

        self.assertEqual(
            first_choice(
                (
                    (
                        "Section 1",
                        (
                            ("first", "First"),
                            ("second", "Second"),
                        ),
                    ),
                ),
            ),
            "first",
        )

        self.assertEqual(
            first_choice(
                (
                    (
                        "Section 1",
                        (),
                    ),
                    (
                        "Section 2",
                        (
                            ("second", "Second"),
                            ("third", "Third"),
                        ),
                    ),
                )
            ),
            "second",
        )

    def test_link_to_framework(self):
        self.assertIsNone(link_to_framework_doc("Card", "this topic does not exist"))
        self.assertIsNone(
            link_to_framework_doc("NonExistingUIITEM", "this topic does not exist")
        )



class JumbotronRenderMixin:
    def render(self, context, instance, placeholder):
        if not getattr(instance, "background_context", False):
            instance.add_classes("bg-light")
        return super().render(context, instance, placeholder)



from django.utils.translation import gettext as _

from djangocms_frontend.models import FrontendUIItem


class Media(FrontendUIItem):
    """
    Layout > "Media" Plugin
    http://getbootstrap.com/docs/4.0/layout/media-object/
    """

    class Meta:
        proxy = True
        verbose_name = _("Media")

    def get_short_description(self):
        return ""


class MediaBody(FrontendUIItem):
    """
    Layout > "Media body" Plugin
    http://getbootstrap.com/docs/4.0/layout/media-object/
    """

    class Meta:
        proxy = True
        verbose_name = _("Media body")

    def get_short_description(self):
        return ""



from django import forms
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin

from djangocms_frontend import settings
from djangocms_frontend.helpers import insert_fields

AUTO_SIZE = (("auto", _("Auto")),)


class SizingMixin:
    def get_fieldsets(self, request, obj=None):
        return insert_fields(
            super().get_fieldsets(request, obj),
            (("size_x", "size_y"),),
            block=0,  # Add to first fieldset - no own tab
            position=-1,
        )

    def render(self, context, instance, placeholder):
        size_x = instance.config.get("size_x", "")
        if size_x.isnumeric() or size_x == "auto":
            instance.add_classes(f"w-{size_x}")
        else:
            instance.add_classes(size_x)
        size_y = instance.config.get("size_y", "")
        if size_y.isnumeric() or size_y == "auto":
            instance.add_classes(f"h-{size_y}")
        else:
            instance.add_classes(size_y)
        return super().render(context, instance, placeholder)


class SizingFormMixin(EntangledModelFormMixin):
    class Meta:
        entangled_fields = {
            "config": [
                "size_x",
                "size_y",
            ]
        }

    size_x = forms.ChoiceField(
        label=_("Horizontal size"),
        required=False,
        initial=settings.EMPTY_CHOICE[0][0],
        choices=settings.EMPTY_CHOICE + settings.SIZE_X_CHOICES,
        #        widget=ButtonGroup(attrs=dict(property="text")),
        help_text=_(
            "Sets the horizontal size relative to the surrounding container or the viewport."
        ),
    )
    size_y = forms.ChoiceField(
        label=_("Vertical size"),
        required=False,
        initial=settings.EMPTY_CHOICE[0][0],
        choices=settings.EMPTY_CHOICE + settings.SIZE_Y_CHOICES,
        #       widget=ButtonGroup(attrs=dict(property="text")),
        help_text=_(
            "Sets the vertical size relative to the surrounding container or the viewport."
        ),
    )



from cms.api import create_page
from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site

DJANGO_CMS4 = apps.is_installed("djangocms_versioning")


class TestFixture:
    """Sets up generic setUp and tearDown methods for tests."""

    def setUp(self):
        self.language = "en"
        self.superuser = self.get_superuser()
        self.default_site = Site.objects.first()
        self.home = self.create_page(
            title="home",
            template="page.html",
        )
        self.publish(self.home, self.language)
        self.page = self.create_page(
            title="content",
            template="page.html",
        )
        self.publish(self.page, self.language)
        self.placeholder = self.get_placeholders(self.page).get(slot="content")
        self.request_url = (
            self.page.get_absolute_url(self.language) + "?toolbar_off=true"
        )
        return super().setUp()

    def tearDown(self):
        self.page.delete()
        self.home.delete()
        if DJANGO_CMS4:
            from djangocms_versioning.models import Version

            Version.objects.all().delete()
        self.superuser.delete()

        return super().tearDown()

    if DJANGO_CMS4:  # CMS V4

        def _get_version(self, grouper, version_state, language=None):
            language = language or self.language

            from djangocms_versioning.models import Version

            versions = Version.objects.filter_by_grouper(grouper).filter(
                state=version_state
            )
            for version in versions:
                if (
                    hasattr(version.content, "language")
                    and version.content.language == language
                ):
                    return version

        def publish(self, grouper, language=None):
            from djangocms_versioning.constants import DRAFT

            version = self._get_version(grouper, DRAFT, language)
            if version is not None:
                version.publish(self.superuser)

        def unpublish(self, grouper, language=None):
            from djangocms_versioning.constants import PUBLISHED

            version = self._get_version(grouper, PUBLISHED, language)
            if version is not None:
                version.unpublish(self.superuser)

        def create_page(self, title, **kwargs):
            kwargs.setdefault("language", self.language)
            kwargs.setdefault("created_by", self.superuser)
            kwargs.setdefault("in_navigation", True)
            kwargs.setdefault("limit_visibility_in_menu", None)
            kwargs.setdefault("menu_title", title)
            return create_page(title=title, **kwargs)

        def get_placeholders(self, page):
            return page.get_placeholders(self.language)

        def create_url(
            self,
            site=None,
            content_object=None,
            manual_url="",
            relative_path="",
            phone="",
            mailto="",
            anchor="",
        ):
            from djangocms_url_manager.models import Url, UrlGrouper
            from djangocms_url_manager.utils import is_versioning_enabled
            from djangocms_versioning.constants import DRAFT
            from djangocms_versioning.models import Version

            if site is None:
                site = self.default_site

            url = Url.objects.create(
                site=site,
                content_object=content_object,
                manual_url=manual_url,
                relative_path=relative_path,
                phone=phone,
                mailto=mailto,
                anchor=anchor,
                url_grouper=UrlGrouper.objects.create(),
            )
            if is_versioning_enabled():
                Version.objects.create(
                    content=url,
                    created_by=self.superuser,
                    state=DRAFT,
                    content_type_id=ContentType.objects.get_for_model(Url).id,
                )

            return url

        def delete_urls(self):
            from djangocms_url_manager.models import Url

            Url.objects.all().delete()

    else:  # CMS V3

        def publish(self, page, language=None):
            page.publish(language)

        def unpublish(self, page, language=None):
            page.unpublish(language)

        def create_page(self, title, **kwargs):
            kwargs.setdefault("language", self.language)
            kwargs.setdefault("menu_title", title)
            return create_page(title=title, **kwargs)

        def get_placeholders(self, page):
            return page.get_placeholders()



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from djangocms_frontend import settings
from djangocms_frontend.common.attributes import AttributesMixin
from djangocms_frontend.common.background import BackgroundMixin
from djangocms_frontend.common.responsive import ResponsiveMixin
from djangocms_frontend.common.sizing import SizingMixin
from djangocms_frontend.common.spacing import SpacingMixin

from ...cms_plugins import CMSUIPlugin
from ...common.title import TitleMixin
from ...helpers import add_plugin
from .. import grid
from . import forms, models

mixin_factory = settings.get_renderer(grid)


@plugin_pool.register_plugin
class GridContainerPlugin(
    mixin_factory("GridContainer"),
    AttributesMixin,
    ResponsiveMixin,
    SpacingMixin,
    BackgroundMixin,
    SizingMixin,
    TitleMixin,
    CMSUIPlugin,
):
    """
    Layout > Grid: "Container" Plugin
    https://getbootstrap.com/docs/5.0/layout/grid/
    """

    name = _("Container")
    module = _("Frontend")
    model = models.GridContainer
    form = forms.GridContainerForm
    change_form_template = "djangocms_frontend/admin/grid_container.html"
    allow_children = True

    fieldsets = [
        (
            None,
            {
                "fields": (
                    (
                        "container_type",
                        "plugin_title",
                    ),
                )
            },
        ),
    ]


@plugin_pool.register_plugin
class GridRowPlugin(
    mixin_factory("GridRow"),
    AttributesMixin,
    ResponsiveMixin,
    SpacingMixin,
    TitleMixin,
    CMSUIPlugin,
):
    """
    Layout > Grid: "Row" Plugin
    https://getbootstrap.com/docs/5.0/layout/grid/
    """

    name = _("Row")
    module = _("Frontend")
    model = models.GridRow
    form = forms.GridRowForm
    change_form_template = "djangocms_frontend/admin/grid_row.html"
    allow_children = True
    child_classes = ["GridColumnPlugin", "CardPlugin"]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    (
                        "create",
                        "plugin_title",
                    ),
                )
            },
        ),
        (
            _("Responsive settings"),
            {
                "fields": ([f"row_cols_{size}" for size in settings.DEVICE_SIZES],),
            },
        ),
        (
            _("Alignment"),
            {
                "fields": (
                    ("vertical_alignment", "horizontal_alignment"),
                    "gutters",
                ),
            },
        ),
    ]

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        data = form.cleaned_data
        for pos in range(data["create"] if data["create"] is not None else 0):
            extra = dict(column_alignment=None)
            for size in settings.DEVICE_SIZES:
                extra[f"{size}_col"] = data.get(f"create_{size}_col")
                extra[f"{size}_order"] = None
                extra[f"{size}_offset"] = None
                extra[f"{size}_ml"] = None
                extra[f"{size}_mr"] = None
            add_plugin(
                obj.placeholder,
                models.GridColumn(
                    parent=obj,
                    placeholder=obj.placeholder,
                    position=obj.position + pos + 1,
                    language=obj.language,
                    plugin_type=GridColumnPlugin.__name__,
                    ui_item=models.GridColumn.__class__.__name__,
                    config=extra,
                ),
            )


@plugin_pool.register_plugin
class GridColumnPlugin(
    mixin_factory("GridColumn"),
    AttributesMixin,
    ResponsiveMixin,
    SpacingMixin,
    BackgroundMixin,
    TitleMixin,
    CMSUIPlugin,
):
    """
    Layout > Grid: "Column" Plugin
    https://getbootstrap.com/docs/5.0/layout/grid/
    """

    name = _("Column")
    module = _("Frontend")
    model = models.GridColumn
    form = forms.GridColumnForm
    change_form_template = "djangocms_frontend/admin/grid_column.html"
    allow_children = True
    require_parent = True
    # TODO it should allow for the responsive utilitiy class
    # https://getbootstrap.com/docs/5.0/layout/grid/#column-resets
    parent_classes = ["GridRowPlugin"]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    (
                        "column_alignment",
                        "text_alignment",
                    ),
                )
            },
        ),
        (
            _("Responsive settings"),
            {
                "fields": (
                    [f"{size}_col" for size in settings.DEVICE_SIZES],
                    [f"{size}_order" for size in settings.DEVICE_SIZES],
                    [f"{size}_offset" for size in settings.DEVICE_SIZES],
                    [f"{size}_ms" for size in settings.DEVICE_SIZES],
                    [f"{size}_me" for size in settings.DEVICE_SIZES],
                )
            },
        ),
        (_("Title settings"), {"fields": ("plugin_title",)}),
    ]



class LinkRenderMixin:
    def render(self, context, instance, placeholder):
        link_classes = []
        if instance.parent and instance.parent.plugin_type == "ListGroupPlugin":
            link_classes.append("list-group-item")
            link_classes.append("list-group-item-action")
            background_prefix = "list-group-item"
        elif (
            getattr(instance, "link_type", "link") == "link"
            and instance.parent
            and instance.parent.plugin_type == "CardInnerPlugin"
        ):
            link_classes.append("card-link")
        else:
            background_prefix = "btn"
        if instance.config.get("link_context", None):
            if getattr(instance, "link_type", "link") == "link":
                link_classes.append(f"link-{instance.link_context}")
            else:
                link_classes.append("btn")
                if not instance.link_outline:
                    link_classes.append(f"{background_prefix}-{instance.link_context}")
                else:
                    link_classes.append(f"btn-outline-{instance.link_context}")
        if instance.config.get("link_size", False):
            link_classes.append(instance.link_size)
        if instance.config.get("link_block", False):
            link_classes.append("d-block")
        if instance.config.get("link_stretched", False):
            link_classes.append("stretched-link")
        context["link"] = instance.get_link()
        instance.add_classes(link_classes)
        return super().render(context, instance, placeholder)



import importlib

from django.conf import settings as django_settings
from django.utils.translation import gettext_lazy as _

EMPTY_CHOICE = (("", "-----"),)

EMPTY_FIELDSET = [
    (
        None,
        {
            "fields": (),
            "description": _(
                "There are no further settings for this plugin. Please press save."
            ),
        },
    )
]

# Only adding block elements
TAG_CHOICES = getattr(
    django_settings,
    "DJANGOCMS_FRONTEND_TAG_CHOICES",
    ["div", "section", "article", "header", "footer", "aside"],
)
TAG_CHOICES = tuple((entry, entry) for entry in TAG_CHOICES)

ADMIN_CSS = getattr(
    django_settings,
    "DJANGOCMS_FRONTEND_ADMIN_CSS",
    {},
)

HEADER_CHOICES = getattr(
    django_settings,
    "DJANGOCMS_FRONTEND_HEADER_CHOICES",
    (
        ("h1", _("Heading 1")),
        ("h2", _("Heading 2")),
        ("h3", _("Heading 3")),
        ("h4", _("Heading 4")),
        ("h5", _("Heading 5")),
    ),
)

ALIGN_CHOICES = (
    ("start", _("Left")),
    ("center", _("Center")),
    ("end", _("Right")),
)

LINK_TEMPLATE_CHOICES = getattr(
    django_settings,
    "DJANGOCMS_FRONTEND_LINK_TEMPLATE_CHOICES",
    [
        ("default", _("Default")),
    ],
)

JUMBOTRON_TEMPLATE_CHOICES = getattr(
    django_settings,
    "DJANGOCMS_FRONTEND_JUMBOTRON_TEMPLATE_CHOICES",
    [
        ("default", _("Default")),
    ],
)

NAVIGATION_TEMPLATE_CHOICES = getattr(
    django_settings,
    "DJANGOCMS_FRONTEND_NAVIGATION_TEMPLATE_CHOICES",
    [
        ("default", _("Default")),
        ("offcanvas", _("Offcanvas")),
    ],
)

FORM_OPTIONS = getattr(django_settings, "DJANGOCMS_FRONTEND_FORM_OPTIONS", {})


framework = getattr(django_settings, "DJANGOCMS_FRONTEND_FRAMEWORK", "bootstrap5")
theme = getattr(django_settings, "DJANGOCMS_FRONTEND_THEME", "djangocms_frontend")

framework_settings = importlib.import_module(
    f"djangocms_frontend.frameworks.{framework}"
)

DEVICE_SIZES = framework_settings.DEVICE_SIZES
DEVICE_CHOICES = framework_settings.DEVICE_CHOICES
COLOR_STYLE_CHOICES = framework_settings.COLOR_STYLE_CHOICES
COLOR_CODES = framework_settings.COLOR_CODES
FORM_TEMPLATE = getattr(framework_settings, "FORM_TEMPLATE", None)
SPACER_PROPERTY_CHOICES = framework_settings.SPACER_PROPERTY_CHOICES
SPACER_SIDE_CHOICES = framework_settings.SPACER_SIDE_CHOICES
SPACER_SIZE_CHOICES = framework_settings.SPACER_SIZE_CHOICES
SPACER_X_SIDES_CHOICES = framework_settings.SPACER_X_SIDES_CHOICES
SPACER_Y_SIDES_CHOICES = framework_settings.SPACER_Y_SIDES_CHOICES
SIZE_X_CHOICES = framework_settings.SIZE_X_CHOICES
SIZE_Y_CHOICES = framework_settings.SIZE_Y_CHOICES
NAVBAR_DESIGNS = framework_settings.NAVBAR_DESIGNS

FRAMEWORK_PLUGIN_INFO = getattr(framework_settings, "FRAMEWORK_PLUGIN_INFO", dict())

EXCL_COL_PROP = FRAMEWORK_PLUGIN_INFO.get("GridColumn", {}).get("excl_col_prop", [])
EXCL_CARD_PROP = FRAMEWORK_PLUGIN_INFO.get("Card", {}).get("excl_card_prop", [])

theme_render_path = f"{theme}.frameworks.{framework}"
theme_forms_path = f"{theme}.forms"


def render_factory(cls, theme_module, render_module):
    parents = tuple(
        getattr(module, cls, None)
        for module in (theme_module, render_module)
        if module is not None and getattr(module, cls, None) is not None
    )
    return type(cls, parents, dict())  # Empty Mix


def get_mixins(naming, theme_path, mixin_path):
    try:
        theme_module = importlib.import_module(theme_path) if theme_path else None
    except ModuleNotFoundError:
        theme_module = None
    try:
        render_module = importlib.import_module(mixin_path) if mixin_path else None
    except ModuleNotFoundError:
        render_module = None

    return lambda name: render_factory(
        naming.format(name=name), theme_module, render_module
    )


def get_renderer(my_module):
    if not isinstance(my_module, str):
        my_module = my_module.__name__
    return get_mixins(
        "{name}RenderMixin", theme_render_path, f"{my_module}.frameworks.{framework}"
    )


def get_forms(my_module):
    if not isinstance(my_module, str):
        my_module = my_module.__name__
    return get_mixins(
        "{name}FormMixin", theme_forms_path, f"{my_module}.frameworks.{framework}"
    )



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from djangocms_frontend.helpers import get_plugin_template

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...common.background import BackgroundMixin
from .. import carousel
from ..link.cms_plugins import LinkPluginMixin
from . import forms, models
from .constants import CAROUSEL_TEMPLATE_CHOICES

mixin_factory = settings.get_renderer(carousel)


@plugin_pool.register_plugin
class CarouselPlugin(mixin_factory("Carousel"), AttributesMixin, CMSUIPlugin):
    """
    Components > "Carousel" Plugin
    https://getbootstrap.com/docs/5.0/components/carousel/
    """

    name = _("Carousel")
    module = _("Frontend")
    model = models.Carousel
    form = forms.CarouselForm
    allow_children = True
    child_classes = ["CarouselSlidePlugin"]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "template",
                    ("carousel_aspect_ratio", "carousel_interval"),
                    ("carousel_controls", "carousel_indicators"),
                    ("carousel_keyboard", "carousel_wrap"),
                    ("carousel_ride",),
                    (
                        "carousel_transition",
                        "carousel_pause",
                    ),
                )
            },
        ),
    ]

    def get_render_template(self, context, instance, placeholder):
        return get_plugin_template(
            instance, "carousel", "carousel", CAROUSEL_TEMPLATE_CHOICES
        )


@plugin_pool.register_plugin
class CarouselSlidePlugin(
    mixin_factory("CarouselSlide"),
    AttributesMixin,
    BackgroundMixin,
    LinkPluginMixin,
    CMSUIPlugin,
):
    """
    Components > "Carousel Slide" Plugin
    https://getbootstrap.com/docs/5.0/components/carousel/
    """

    name = _("Carousel slide")
    module = _("Frontend")
    model = models.CarouselSlide
    form = forms.CarouselSlideForm
    allow_children = True
    parent_classes = ["CarouselPlugin"]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "carousel_image",
                    "carousel_content",
                )
            },
        ),
        (
            _("Link settings"),
            {
                "classes": ("collapse",),
                "fields": (
                    ("external_link", "internal_link"),
                    ("mailto", "phone"),
                    ("anchor", "target"),
                ),
            },
        ),
    ]

    def get_render_template(self, context, instance, placeholder):
        return get_plugin_template(
            instance.parent.get_plugin_instance()[0],
            "carousel",
            "slide",
            CAROUSEL_TEMPLATE_CHOICES,
        )



class MediaRenderMixin:
    render_template = "djangocms_frontend/bootstrap5/media.html"

    def render(self, context, instance, placeholder):
        instance.add_classes("d-flex")
        return super().render(context, instance, placeholder)


class MediaBodyRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes("flex-grow-1", "ms-3")
        return super().render(context, instance, placeholder)



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...common.background import BackgroundMixin
from ...common.responsive import ResponsiveMixin
from ...common.spacing import SpacingMixin
from ...helpers import get_plugin_template
from .. import jumbotron
from . import forms, models

mixin_factory = settings.get_renderer(jumbotron)


@plugin_pool.register_plugin
class JumbotronPlugin(
    mixin_factory("Jumbotron"),
    AttributesMixin,
    ResponsiveMixin,
    SpacingMixin,
    BackgroundMixin,
    CMSUIPlugin,
):
    """
    Components > "Jumbotron" Plugin
    https://getbootstrap.com/docs/5.1/examples/jumbotron/
    """

    name = _("Jumbotron")
    module = _("Frontend")
    model = models.Jumbotron
    form = forms.JumbotronForm
    change_form_template = "djangocms_frontend/admin/jumbotron.html"
    allow_children = True

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "template",
                    "jumbotron_fluid",
                )
            },
        ),
    ]

    def get_render_template(self, context, instance, placeholder):
        return get_plugin_template(
            instance, "jumbotron", "jumbotron", settings.JUMBOTRON_TEMPLATE_CHOICES
        )



from django.test import TestCase

from djangocms_frontend.contrib.tabs.models import Tab, TabItem


class TabsModelTestCase(TestCase):
    def test_tab_instance(self):
        instance = Tab.objects.create(
            config=dict(tab_type="nav-tabs", tab_alignment="")
        )
        self.assertEqual(str(instance), "Tab (1)")
        self.assertEqual(instance.get_short_description(), "(nav-tabs)")
        instance.config["tab_alignment"] = "nav-fill"
        self.assertEqual(instance.get_short_description(), "(nav-tabs) .nav-fill")

    def test_tab_item_instance(self):
        instance = TabItem.objects.create(config=dict(tab_title=""))
        self.assertEqual(str(instance), "TabItem (1)")
        self.assertEqual(instance.get_short_description(), "")



class CollapseRenderMixin:
    render_template = "djangocms_frontend/bootstrap5/collapse.html"


class CollapseContainerRenderMixin:
    render_template = "djangocms_frontend/bootstrap5/collapse-container.html"

    def render(self, context, instance, placeholder):
        instance.add_classes("collapse")
        return super().render(context, instance, placeholder)


class CollapseTriggerRenderMixin:
    render_template = "djangocms_frontend/bootstrap5/collapse-trigger.html"



from django.utils.translation import gettext_lazy as _

CODE_TYPE_CHOICES = (
    ("code", _("Inline code")),
    ("pre", _("Code block")),
    ("var", _("Variables")),
    ("kbd", _("User input")),
    ("samp", _("Sample output")),
)



import djangocms_frontend

from .subcommands.base import SubcommandsCommand
from .subcommands.frequency_analysis import FrequencyAnalysis
from .subcommands.migrate import Migrate
from .subcommands.stale_references import StaleReferences
from .subcommands.sync_permissions import SyncPermissions


class Command(SubcommandsCommand):
    command_name = "frontend"
    subcommands = {
        "migrate": Migrate,
        "frequency_analysis": FrequencyAnalysis,
        "stale_references": StaleReferences,
        "sync_permissions": SyncPermissions,
    }
    missing_args_message = "one of the available sub commands must be provided"

    subcommand_dest = "command"

    def get_version(self):
        return djangocms_frontend.__version__

    def add_arguments(self, parser):
        parser.add_argument("--version", action="version", version=self.get_version())
        super().add_arguments(parser)



from django.test import TestCase

from djangocms_frontend.fields import AttributesField, TagTypeField, TagTypeFormField


class FieldsTestCase(TestCase):
    def test_attributes_field(self):
        field = AttributesField()
        self.assertEqual(field.verbose_name, "Attributes")
        self.assertEqual(field.blank, True)

    def test_tag_type_field(self):
        field = TagTypeField()
        self.assertEqual(field.verbose_name, "Tag type")
        self.assertEqual(field.default, "div")
        self.assertEqual(field.max_length, 255)
        self.assertEqual(
            field.help_text,
            "Select the HTML tag to be used.",
        )

        form_field = TagTypeFormField()
        self.assertEqual(
            form_field.choices,
            [
                ("div", "div"),
                ("section", "section"),
                ("article", "article"),
                ("header", "header"),
                ("footer", "footer"),
                ("aside", "aside"),
            ],
        )



import json

from django import template
from django.contrib.contenttypes.models import ContentType
from django.core.serializers.json import DjangoJSONEncoder
from django.template.defaultfilters import safe
from django.utils.encoding import force_str
from django.utils.functional import Promise
from django.utils.html import conditional_escape, mark_safe

from djangocms_frontend import settings
from djangocms_frontend.fields import HTMLsanitized
from djangocms_frontend.helpers import get_related_object as related_object

register = template.Library()


@register.simple_tag
def get_attributes(attribute_field, *add_classes):
    """Joins a list of classes with an attributes field and returns all html attributes"""
    additional_classes = set()
    for classes in add_classes:
        if classes:
            additional_classes.update(
                classes.split() if isinstance(classes, str) else classes
            )
    attrs = []
    if attribute_field:
        for key, val in attribute_field.items():
            if key.lower() == "class":
                val = " ".join(additional_classes.union(set(val.split())))
            if val:
                attrs.append(f'{key}="{conditional_escape(val)}"')
            else:
                attrs.append(f"{key}")
    if additional_classes and (not attribute_field or "class" not in attribute_field):
        attrs.append(f'class="{conditional_escape(" ".join(additional_classes))}"')
    return mark_safe(" ".join(attrs))


@register.filter
def get_related_object(reference):
    return related_object(dict(obj=reference), "obj")


class LazyEncoder(DjangoJSONEncoder):
    def default(self, obj):
        if isinstance(obj, Promise):
            return force_str(obj)
        return super().default(obj)


@register.filter
def json_dumps(data):
    """Converts data to JSON forcing text on lazy gettext translation"""
    return mark_safe(json.dumps(data, cls=LazyEncoder))


@register.filter
def html_safe(data):
    if HTMLsanitized:
        return safe(data)
    return data


@register.filter
def safe_caption(data):
    if data[:3] == "<p>" and data[-4:] == "</p>":
        block = data[3:-4]
        if "<p>" not in block:
            data = block
    if HTMLsanitized:
        return safe(data)
    return data


@register.simple_tag(takes_context=True)
def framework_info(context, item, as_json=True):
    """Retrieves framework_info for plugin either from existing plugin or from
    content type"""
    plugin = context.get("plugin", None)
    if plugin:  # if possible get from plugin
        return (
            mark_safe(json.dumps(plugin.framework_info.get(item, ""), cls=LazyEncoder))
            if as_json
            else plugin.get(item, "")
        )
    framework_info = context.get("framework_info", {})  # already available
    if not framework_info:
        content_type_id = context.get("content_type_id", None)
        if content_type_id:  # Get from content_type
            model_name = (
                ContentType.objects.get(id=content_type_id).model_class().__name__
            )
            framework_info = settings.FRAMEWORK_PLUGIN_INFO.get(model_name, {})
            context["framework_info"] = framework_info  # and store
    return (
        mark_safe(json.dumps(framework_info.get(item, ""), cls=LazyEncoder))
        if as_json
        else framework_info.get(item, "")
    )



from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy

DEVICE_CHOICES = (
    ("xs", _("Extra small")),  # default <576px
    ("sm", _("Small")),  # default ≥576px
    ("md", _("Medium")),  # default ≥768px
    ("lg", _("Large")),  # default ≥992px
    ("xl", _("Extra large")),  # default ≥1200px
    ("xxl", _("Extra-extra large")),  # default ≥1200px
)
DEVICE_SIZES = tuple(size for size, name in DEVICE_CHOICES)

COLOR_STYLE_CHOICES = getattr(
    settings,
    "DJANGOCMS_FRONTEND_COLOR_STYLE_CHOICES",
    (
        ("primary", _("Primary")),
        ("secondary", _("Secondary")),
        ("success", _("Success")),
        ("danger", _("Danger")),
        ("warning", _("Warning")),
        ("info", _("Info")),
        ("light", _("Light")),
        ("dark", _("Dark")),
    ),
)

COLOR_CODES = getattr(
    settings,
    "DJANGOCMS_FRONTEND_COLOR_CODES",
    dict(),
)

FORM_TEMPLATE = getattr(
    settings,
    "DJANGOCMS_FRONTEND_FORM_TEMPLATE",
    "djangocms_frontend/bootstrap5/render/form.html",
)

SPACER_PROPERTY_CHOICES = (
    ("m", "margin"),
    ("p", "padding"),
)

SPACER_SIDE_CHOICES = (
    ("", "*"),
    ("t", "*-top"),
    ("r", "*-right"),
    ("b", "*-bottom"),
    ("l", "*-left"),
    ("x", "*-left & *-right"),
    ("y", "*-top & *-bottom"),
)

SPACER_X_SIDES_CHOICES = (
    ("x", _("Both")),
    ("s", _("Left")),
    ("e", _("Right")),
)

SPACER_Y_SIDES_CHOICES = (
    ("y", _("Both")),
    ("t", _("Top")),
    ("b", _("Bottom")),
)

SPACER_SIZE_CHOICES = getattr(
    settings,
    "DJANGOCMS_FRONTEND_SPACER_SIZES",
    (
        ("0", "* 0"),
        ("1", "* .25"),
        ("2", "* .5"),
        ("3", "* 1"),
        ("4", "* 1.5"),
        ("5", "* 3"),
    ),
)

SIZE_X_CHOICES = getattr(
    settings,
    "DJANGOCMS_FRONTEND_SIZE_X_CHOICES",
    (
        ("25", "25%"),
        ("50", "50%"),
        ("75", "75%"),
        ("100", "100%"),
        ("auto", _("Auto")),
        ("vw-100", _("Screen")),
    ),
)

SIZE_Y_CHOICES = getattr(
    settings,
    "DJANGOCMS_FRONTEND_SIZE_Y_CHOICES",
    (
        ("25", "25%"),
        ("50", "50%"),
        ("75", "75%"),
        ("100", "100%"),
        ("auto", _("Auto")),
        ("min-vh-100", _("Screen (minimum)")),
    ),
)

OPACITY_CHOICES = getattr(
    settings,
    "DJANGOCMS_FRONTEND_OPACITY_CHOICES",
    (
        (100, "100%"),
        (75, "75%"),
        (50, "50%"),
        (25, "25%"),
        (10, "10%"),
    ),
)

SHADOW_CHOICES = getattr(
    settings,
    "DJANGOCMS_FRONTEND_SHADOW_CHOICES",
    (
        ("none", pgettext_lazy("shadow", "None")),
        ("sm", "S"),
        ("reg", "M"),
        ("lg", "L"),
    ),
)

grid_sizes = [
    _("Extra small"),
    _("Small"),
    _("Medium"),
    _("Large"),
    _("Extra large"),
    _("XX large"),
]

grid_icons = ["size-xs", "size-sm", "size-md", "size-lg", "size-xl", "size-xxl"]


FRAMEWORK_PLUGIN_INFO = {
    "GridColumn": {
        "grid_sizes": grid_sizes,
        "grid_icons": grid_icons,
        "row_links": [
            "https://getbootstrap.com/docs/5.1/layout/grid/#grid-options",
            "https://getbootstrap.com/docs/5.1/layout/columns/#reordering",
            "https://getbootstrap.com/docs/5.1/layout/columns/#offsetting-columns",
            "https://getbootstrap.com/docs/5.1/utilities/flex/#auto-margins",
            "https://getbootstrap.com/docs/5.1/utilities/flex/#auto-margins",
        ],
    },
    "GridRow": {
        "grid_sizes": grid_sizes,
        "grid_icons": grid_icons,
        "row_links": [
            "https://getbootstrap.com/docs/5.1/layout/grid/#row-columns",
        ],
        "vertical_alignment_link": "https://getbootstrap.com/docs/5.1/layout/columns/#vertical-alignment",
        "horizontal_alignment_link": "https://getbootstrap.com/docs/5.1/layout/columns/#horizontal-alignment",
    },
    "CardLayout": {
        "card_type_link": "https://getbootstrap.com/docs/5.1/components/card/#card-layout",
        "grid_sizes": grid_sizes,
        "grid_icons": grid_icons,
        "row_links": [
            "https://getbootstrap.com/docs/5.1/layout/grid/#row-columns",
        ],
    },
}

NAVBAR_DESIGNS = getattr(
    settings,
    "DJANGOCMS_FRONTEND_NAVBAR_DESIGN",
    (
        ("light", _("Light")),
        ("dark", _("Dark")),
    ),
)



from django import forms
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from djangocms_attributes_field import fields

from . import settings
from .helpers import first_choice


class TemplateChoiceMixin:
    """Mixin that hides the template field if only one template is available and is selected"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "template" in self.fields:
            template_field = self.fields["template"]
            choices = template_field.choices
            instance = kwargs.get("instance", None)
            if len(choices) == 1 and (
                instance is None or instance.config.get("template", "") == choices[0][0]
            ):
                template_field.widget = forms.HiddenInput()


class ButtonGroup(forms.RadioSelect):
    template_name = "djangocms_frontend/admin/widgets/button_group.html"
    option_template_name = "djangocms_frontend/admin/widgets/button_group_option.html"

    class Media:
        css = {"all": ("djangocms_frontend/css/button_group.css",)}


class ColoredButtonGroup(ButtonGroup):  # lgtm [py/missing-call-to-init]
    option_template_name = (
        "djangocms_frontend/admin/widgets/button_group_color_option.html"
    )

    class Media:
        css = settings.ADMIN_CSS

    def __init__(self, *args, **kwargs):
        kwargs.update({"attrs": {**kwargs.get("attrs", {}), **dict(property="color")}})
        super().__init__(*args, **kwargs)


class IconGroup(ButtonGroup):  # lgtm [py/missing-call-to-init]
    option_template_name = "djangocms_frontend/admin/widgets/icon_group_option.html"

    def __init__(self, *args, **kwargs):
        kwargs.update({"attrs": {**dict(property="icon"), **kwargs.get("attrs", {})}})
        super().__init__(*args, **kwargs)


class IconMultiselect(forms.CheckboxSelectMultiple):  # lgtm [py/missing-call-to-init]
    template_name = "djangocms_frontend/admin/widgets/button_group.html"
    option_template_name = "djangocms_frontend/admin/widgets/icon_group_option.html"

    class Media:
        css = {"all": ("djangocms_frontend/css/button_group.css",)}

    def __init__(self, *args, **kwargs):
        kwargs.update({"attrs": {**kwargs.get("attrs", {}), **dict(property="icon")}})
        super().__init__(*args, **kwargs)


class OptionalDeviceChoiceField(
    forms.MultipleChoiceField
):  # lgtm [py/missing-call-to-init]
    def __init__(self, **kwargs):
        kwargs.setdefault("choices", settings.DEVICE_CHOICES)
        kwargs.setdefault("initial", None)
        kwargs.setdefault("widget", IconMultiselect())
        super().__init__(**kwargs)

    def prepare_value(self, value):
        if value is None:
            value = [size for size, _ in settings.DEVICE_CHOICES]
        return super().prepare_value(value)

    def clean(self, value):
        value = super().clean(value)
        if len(value) == len(settings.DEVICE_CHOICES):
            return None
        return value


class DeviceChoiceField(OptionalDeviceChoiceField):
    def clean(self, value):
        value = super().clean(value)
        if isinstance(value, list) and len(value) == 0:
            raise ValidationError(
                _("Please select at least one device size"), code="invalid"
            )
        return value


class AttributesField(fields.AttributesField):
    def __init__(self, *args, **kwargs):
        if "verbose_name" not in kwargs:
            kwargs["verbose_name"] = _("Attributes")
        if "blank" not in kwargs:
            kwargs["blank"] = True
        super().__init__(*args, **kwargs)


class AttributesFormField(fields.AttributesFormField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("label", _("Attributes"))
        kwargs.setdefault("required", False)
        kwargs.setdefault("widget", fields.AttributesWidget)
        self.excluded_keys = kwargs.pop("excluded_keys", [])
        super().__init__(*args, **kwargs)


try:
    fields.AttributesWidget(
        sorted=True
    )  # does djangocms-attributes-field support sorted param?
    CHOICESWIDGETPARAMS = dict(sorted=False)  # use unsorted variant
except TypeError:
    CHOICESWIDGETPARAMS = dict()  # Fallback for djangocms-attributes-field < 2.1


class ChoicesFormField(fields.AttributesFormField):
    """Simple choices field based on attributes field. Needs to be extended to
    allow to sort choices"""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("label", _("Choices"))
        kwargs.setdefault("required", True)
        kwargs.setdefault("widget", fields.AttributesWidget(**CHOICESWIDGETPARAMS))
        self.excluded_keys = kwargs.pop("excluded_keys", [])
        super().__init__(*args, **kwargs)

    def clean(self, value):
        if not value:
            raise ValidationError(
                mark_safe(
                    _(
                        "Please enter at least one choice. Use the <code>+</code> symbol to add a choice."
                    )
                ),
                code="empty",
            )
        return [(key, value) for key, value in value.items()]

    def prepare_value(self, value):
        if not value:
            return {}
        if isinstance(value, dict):  # Already dict? OK!
            return super().prepare_value(value)
        # Turn items into dict
        return super().prepare_value({key: value for key, value in value})


class TagTypeField(models.CharField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("verbose_name", _("Tag type"))
        # Choices are not defined at database level but in the form field below
        kwargs.setdefault("default", first_choice(settings.TAG_CHOICES))
        kwargs.setdefault("max_length", 255)
        kwargs.setdefault("help_text", _("Select the HTML tag to be used."))
        super().__init__(*args, **kwargs)


class TagTypeFormField(forms.ChoiceField):  # lgtm [py/missing-call-to-init]
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("label", _("Tag type"))
        kwargs.setdefault("choices", settings.TAG_CHOICES)
        kwargs.setdefault("initial", first_choice(settings.TAG_CHOICES))
        kwargs.setdefault("required", False)
        kwargs.setdefault("widget", ButtonGroup(attrs=dict(property="text")))
        super().__init__(*args, **kwargs)


class AutoNumberInput(forms.NumberInput):  # lgtm [py/missing-call-to-init]
    class Media:
        js = ("djangocms_frontend/js/auto_input.js",)

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("attrs", {"class": "auto-field"})
        super().__init__(*args, **kwargs)


try:
    from djangocms_text_ckeditor.fields import HTMLFormField  # noqa

    HTMLsanitized = True
except ModuleNotFoundError:
    HTMLFormField = forms.CharField
    HTMLsanitized = False



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from djangocms_frontend.helpers import get_plugin_template

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...common.spacing import PaddingMixin
from .. import tabs
from . import forms, models
from .constants import TAB_TEMPLATE_CHOICES

mixin_factory = settings.get_renderer(tabs)


@plugin_pool.register_plugin
class TabPlugin(mixin_factory("Tab"), AttributesMixin, CMSUIPlugin):
    """
    Components > "Navs - Tab" Plugin
    https://getbootstrap.com/docs/5.0/components/navs/
    """

    name = _("Tabs")
    module = _("Frontend")
    model = models.Tab
    form = forms.TabForm
    change_form_template = "djangocms_frontend/admin/tabs.html"
    allow_children = True
    child_classes = ["TabItemPlugin"]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "template",
                    ("tab_type", "tab_alignment"),
                    ("tab_index", "tab_effect"),
                )
            },
        ),
    ]

    def get_render_template(self, context, instance, placeholder):
        return get_plugin_template(instance, "tabs", "tabs", TAB_TEMPLATE_CHOICES)


@plugin_pool.register_plugin
class TabItemPlugin(
    mixin_factory("TabItem"), AttributesMixin, PaddingMixin, CMSUIPlugin
):
    """
    Components > "Navs - Tab Item" Plugin
    https://getbootstrap.com/docs/5.0/components/navs/
    """

    name = _("Tab item")
    module = _("Frontend")
    model = models.TabItem
    form = forms.TabItemForm
    change_form_template = "djangocms_frontend/admin/tabs.html"
    allow_children = True
    parent_classes = ["TabPlugin"]

    fieldsets = [
        (
            None,
            {"fields": ("tab_title", "tab_bordered")},
        ),
    ]

    def get_render_template(self, context, instance, placeholder):
        return get_plugin_template(
            instance.parent.get_plugin_instance()[0],
            "tabs",
            "item",
            TAB_TEMPLATE_CHOICES,
        )



from django.utils.translation import gettext_lazy as _

from djangocms_frontend.models import FrontendUIItem
from djangocms_frontend.settings import COLOR_STYLE_CHOICES

# card allow for a transparent color
CARD_COLOR_STYLE_CHOICES = COLOR_STYLE_CHOICES + (("transparent", _("Transparent")),)

CARD_TEXT_STYLES = COLOR_STYLE_CHOICES + (("white", _("White")),)


class CardLayout(FrontendUIItem):
    """
    Components > "Card" Plugin
    https://getbootstrap.com/docs/5.0/components/card/
    """

    class Meta:
        proxy = True
        verbose_name = _("Card layout")

    def get_short_description(self):
        text = f"({self.card_type})"
        return text


class Card(FrontendUIItem):
    """
    Components > "Card" Plugin
    https://getbootstrap.com/docs/5.0/components/card/
    """

    class Meta:
        proxy = True
        verbose_name = _("Card")

    def get_short_description(self):
        if self.config.get("background_context", False):
            text = [f".bg-{self.background_context}"]
        else:
            text = []
        if self.config.get("card_outline", False):
            text.append(f".border-{self.card_outline}")
        if self.card_alignment:
            text.append(f".{self.card_alignment}")
        return " ".join(text)


class CardInner(FrontendUIItem):
    """
    Components > "Card - Inner" Plugin (Header, Footer, Body)
    https://getbootstrap.com/docs/5.0/components/card/
    """

    class Meta:
        proxy = True
        verbose_name = _("Card inner")

    def get_short_description(self):
        return "({})".format(self.inner_type.split("-", 1)[-1])



from django import forms
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelForm

from djangocms_frontend import settings
from djangocms_frontend.common.responsive import ResponsiveFormMixin
from djangocms_frontend.common.spacing import SpacingFormMixin
from djangocms_frontend.contrib import alert
from djangocms_frontend.fields import (
    AttributesFormField,
    ColoredButtonGroup,
    TagTypeFormField,
)
from djangocms_frontend.helpers import first_choice
from djangocms_frontend.models import FrontendUIItem
from djangocms_frontend.settings import COLOR_STYLE_CHOICES

mixin_factory = settings.get_forms(alert)


class AlertForm(
    mixin_factory("Alert"), ResponsiveFormMixin, SpacingFormMixin, EntangledModelForm
):
    """
    Components > "Alerts" Plugin
    https://getbootstrap.com/docs/5.0/components/alerts/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "alert_context",
                "alert_dismissible",
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    alert_context = forms.ChoiceField(
        label=_("Context"),
        choices=COLOR_STYLE_CHOICES,
        initial=first_choice(COLOR_STYLE_CHOICES),
        widget=ColoredButtonGroup(),
    )
    alert_dismissible = forms.BooleanField(
        label=_("Dismissible"),
        initial=False,
        required=False,
        help_text=_("Allows the alert to be closed."),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()



from cms.api import add_plugin
from cms.test_utils.testcases import CMSTestCase

from djangocms_frontend.contrib.media.cms_plugins import MediaBodyPlugin, MediaPlugin

from ..fixtures import TestFixture


class MediaPluginTestCase(TestFixture, CMSTestCase):
    def test_media_plugin(self):
        add_plugin(
            placeholder=self.placeholder,
            plugin_type=MediaPlugin.__name__,
            language=self.language,
        )
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<div class="flex-shrink-0">')

    def test_media_body_plugin(self):
        add_plugin(
            placeholder=self.placeholder,
            plugin_type=MediaBodyPlugin.__name__,
            language=self.language,
        )
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "flex-grow-1")



from django import forms
from django.conf import settings as django_settings
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelForm

from djangocms_frontend.settings import ALIGN_CHOICES

from ... import settings
from ...common.background import BackgroundFormMixin
from ...common.responsive import ResponsiveFormMixin
from ...common.spacing import SpacingFormMixin
from ...fields import AttributesFormField, HTMLFormField, IconGroup, TagTypeFormField
from ...helpers import first_choice
from ...models import FrontendUIItem
from .. import content
from .constants import CODE_TYPE_CHOICES

mixin_factory = settings.get_forms(content)


class CodeForm(
    mixin_factory("Code"),
    SpacingFormMixin,
    ResponsiveFormMixin,
    BackgroundFormMixin,
    EntangledModelForm,
):
    """
    Content > "Code" Plugin
    https://getbootstrap.com/docs/5.0/content/code/
    """

    class Media:
        js = (
            "admin/vendor/ace/ace.js"
            if "djangocms_static_ace" in django_settings.INSTALLED_APPS
            else "https://cdnjs.cloudflare.com/ajax/libs/ace/1.9.6/ace.js",
        )

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "code_content",
                "code_type",
                "attributes",
            ]
        }

    code_content = forms.CharField(
        label=_("Code"),
        initial="",
        required=True,
        widget=forms.widgets.Textarea(attrs={"class": "js-ckeditor-use-selected-text"}),
    )
    code_type = forms.ChoiceField(
        label=_("Code type"),
        choices=CODE_TYPE_CHOICES,
        initial=first_choice(CODE_TYPE_CHOICES),
        required=True,
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()


class BlockquoteForm(
    mixin_factory("Blockquote"),
    SpacingFormMixin,
    ResponsiveFormMixin,
    BackgroundFormMixin,
    EntangledModelForm,
):
    """
    Content > "Blockquote" Plugin
    https://getbootstrap.com/docs/5.0/content/typography/#blockquotes
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "quote_content",
                "quote_origin",
                "quote_alignment",
                "attributes",
            ]
        }

    quote_content = HTMLFormField(
        label=_("Quote"),
        initial="",
        required=True,
    )
    quote_origin = HTMLFormField(
        label=_("Source"),
        required=False,
    )
    quote_alignment = forms.ChoiceField(
        label=_("Alignment"),
        choices=settings.EMPTY_CHOICE + ALIGN_CHOICES,
        initial=settings.EMPTY_CHOICE[0][0],
        required=False,
        widget=IconGroup(),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()


class FigureForm(
    mixin_factory("Figure"),
    SpacingFormMixin,
    ResponsiveFormMixin,
    BackgroundFormMixin,
    EntangledModelForm,
):
    """
    Content > "Figure" Plugin
    https://getbootstrap.com/docs/5.0/content/figures/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "figure_caption",
                "figure_alignment",
                "attributes",
            ]
        }

    figure_caption = HTMLFormField(
        label=_("Caption"),
        initial="",
        required=True,
    )
    figure_alignment = forms.ChoiceField(
        label=_("Alignment"),
        choices=settings.EMPTY_CHOICE + ALIGN_CHOICES,
        initial=settings.EMPTY_CHOICE[0][0],
        required=False,
        widget=IconGroup(),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()



from django.conf import settings as django_setting
from djangocms_bootstrap4.constants import DEVICE_SIZES

from djangocms_frontend import settings


def breakpoints(props):
    lst = []
    for size in DEVICE_SIZES:
        for prop in props:
            if prop == "ml":
                lst.append(f"{size}_ml -> {size}_ms")
            elif prop == "mr":
                lst.append(f"{size}_mr -> {size}_me")
            else:
                lst.append(f"{size}_{prop}")
    return lst


plugin_migrations = {
    "bootstrap4_alerts.Bootstrap4Alerts -> alert.Alert": [
        "alert_context",
        "alert_dismissable -> alert_dismissible",
        "tag_type",
        "attributes",
        "P001",  # additional data migration, see below
        "M001-m",  # SpacingMixin
        "M001-p",  # SpacingMixin
        "M002",  # ResponsiveMixin
    ],
    "bootstrap4_badge.Bootstrap4Badge -> badge.Badge": [
        "badge_text",
        "badge_context",
        "badge_pills",
        "attributes",
        "P001",  # additional data migration, see below
    ],
    "bootstrap4_card.Bootstrap4Card -> card.Card": [
        "card_type",
        "card_context -> background_context",
        "card_alignment",
        "card_outline",
        "card_text_color",
        "tag_type",
        "attributes",
        "P001",
        "X002",  # Replace v4 card deck
        "X003",  # Align card_outline and background_context
        "A001_card",  # fix alignment
        "M001-m",  # SpacingMixin
        "M002",  # ResponsiveMixin
        "M003",  # BackgroundMixin
    ],
    "bootstrap4_card.Bootstrap4CardInner -> card.CardInner": [
        "inner_type",
        "tag_type",
        "attributes",
        "M001-p",  # SpacingMixin
        "M002",  # ResponsiveMixin
        "M003",  # BackgroundMixin
    ],
    # collapse
    "bootstrap4_collapse.Bootstrap4Collapse -> collapse.Collapse": [
        "siblings",
        "tag_type",
        "attributes",
        "P001",
    ],
    "bootstrap4_collapse.Bootstrap4CollapseContainer -> collapse.CollapseContainer": [
        "identifier -> container_identifier",
        "tag_type",
        "attributes",
    ],
    "bootstrap4_collapse.Bootstrap4CollapseTrigger -> collapse.CollapseTrigger": [
        "identifier -> trigger_identifier",
        "tag_type",
        "attributes",
        "P001",
    ],
    "bootstrap4_content.Bootstrap4Blockquote -> content.Blockquote": [
        "quote_content",
        "quote_origin",
        "quote_alignment",
        "attributes",
        "P001",
        "A001_quote",  # fix alignment
        "M001-m",  # SpacingMixin
        "M001-p",  # SpacingMixin
        "M002",  # ResponsiveMixin
        "M003",  # BackgroundMixin
    ],
    "bootstrap4_content.Bootstrap4Code -> content.CodeBlock": [
        "code_content",
        "attributes",
        "P001",
        "M001-m",  # SpacingMixin
        "M001-p",  # SpacingMixin
        "M002",  # ResponsiveMixin
        "M003",  # BackgroundMixin
    ],
    "bootstrap4_content.Bootstrap4Figure -> content.Figure": [
        "figure_caption",
        "figure_alignment",
        "attributes",
        "M001-m",  # SpacingMixin
        "M001-p",  # SpacingMixin
        "M002",  # ResponsiveMixin
        "M003",  # BackgroundMixin
    ],
    "bootstrap4_grid.Bootstrap4GridContainer -> grid.GridContainer": [
        "container_type",
        "attributes",
        "P001",
        "tag_type",
        "M001-m",  # SpacingMixin
        "M001-p",  # SpacingMixin
        "M002",  # ResponsiveMixin
        "M003",  # BackgroundMixin
    ],
    "bootstrap4_grid.Bootstrap4GridRow -> grid.GridRow": [
        "horizontal_alignment",
        "vertical_alignment",
        "gutters",
        "attributes",
        "P001",
        "tag_type",
        "M001-m",  # SpacingMixin
        "M001-p",  # SpacingMixin
        "M002",  # ResponsiveMixin
        "M003",  # BackgroundMixin
    ],
    "bootstrap4_grid.Bootstrap4GridColumn -> grid.GridColumn": [
        "column_alignment",
        "() -> text_alignment",
        "attributes",
        "P001",
        "G001",  # fill text_alignment from attributes if possible
        "tag_type",
        "M001-m",  # SpacingMixin
        "M001-p",  # SpacingMixin
        "M002",  # ResponsiveMixin
        "M003",  # BackgroundMixin
    ]
    + breakpoints(("col", "order", "ml", "mr", "offset")),
    "bootstrap4_jumbotron.Bootstrap4Jumbotron -> jumbotron.Jumbotron": [
        "fluid -> jumbotron_fluid",
        "(default) -> template",
        "() -> jumbotron_context",
        "() -> jumbotron_opacity",
        "tag_type",
        "attributes",
        "P001",
        "M001-m",  # SpacingMixin
        "M001-p",  # SpacingMixin
        "M002",  # ResponsiveMixin
        "M003",  # BackgroundMixin
    ],
    "bootstrap4_link.Bootstrap4Link -> link.Link": [
        "template",
        "name",
        "external_link",
        "anchor",
        "mailto",
        "phone",
        "link_type",
        "target -> link_target",
        "link_context",
        "link_size",
        "link_outline",
        "link_block",
        "internal_link",
        "icon_left",
        "icon_right",
        "file_link",
        "attributes",
        "P001",
        "M001-m",  # SpacingMixin
        "M001-p",  # SpacingMixin
        "T001_LINK",
    ],
    "bootstrap4_listgroup.Bootstrap4ListGroup -> listgroup.ListGroup": [
        "list_group_flush",
        "tag_type",
        "attributes",
        "P001",
        "M001-m",  # SpacingMixin
        "M002",  # ResponsiveMixin
    ],
    "bootstrap4_listgroup.Bootstrap4ListGroupItem -> listgroup.ListGroupItem": [
        "list_context",
        "list_state",
        "tag_type",
        "attributes",
        "P001",
        "M001-p",  # SpacingMixin
    ],
    "bootstrap4_media.Bootstrap4Media -> media.Media": [
        "tag_type",
        "attributes",
        "M002",  # ResponsiveMixin
    ],
    "bootstrap4_media.Bootstrap4MediaBody -> media.MediaBody": [
        "tag_type",
        "attributes",
    ],
    "bootstrap4_picture.Bootstrap4Picture -> image.Image": [
        "template",
        "external_picture",
        "width",
        "height",
        "alignment",
        "caption_text",
        "link_url -> external_link",
        "link_target",
        "link_attributes",
        "use_automatic_scaling",
        "use_no_cropping",
        "use_crop",
        "use_upscale",
        "picture_fluid",
        "picture_rounded",
        "picture_thumbnail",
        "link_page -> internal_link",
        "picture",
        "thumbnail_options",
        "use_responsive_image",
        "attributes",
        "P001",
        "A001_picture",  # fix alignment
        "M001-m",  # MarginMixin
        "M002",  # ResponsiveMixin
        "T001_PICTURE",
    ],
    "bootstrap4_tabs.Bootstrap4Tab -> tabs.Tab": [
        "template",
        "tab_type",
        "tab_alignment",
        "tab_index",
        "tab_effect",
        "tag_type",
        "attributes",
        "P001",
        "T001_TABS",
    ],
    "bootstrap4_tabs.Bootstrap4TabItem -> tabs.TabItem": [
        "tab_title",
        "tag_type",
        "attributes",
        "P001",
        "M001-p",  # SpacingMixin
    ],
    "bootstrap4_utilities.Bootstrap4Spacing -> utilities.Spacing": [
        "space_property",
        "space_sides",
        "space_device",
        "space_size",
        "tag_type",
        "attributes",
        "P001",
    ],
    "bootstrap4_carousel.Bootstrap4Carousel -> carousel.Carousel": [
        "template",
        "carousel_interval",
        "carousel_controls",
        "carousel_indicators",
        "carousel_keyboard",
        "carousel_pause",
        "carousel_ride",
        "carousel_wrap",
        "carousel_aspect_ratio",
        "tag_type",
        "attributes",
        "P001",
        "T001_CAROUSEL",
    ],
    "bootstrap4_carousel.Bootstrap4CarouselSlide -> carousel.CarouselSlide": [
        "template",
        "carousel_image",
        "carousel_content",
        "external_link",
        "target",
        "internal_link",
        "file_link",
        "tag_type",
        "attributes",
        "P001",
        "T001_CAROUSEL_SLIDE",
    ],
}


def p001_left_right_migration(obj, new_obj):
    if "class" in new_obj.attributes:

        def replace(item, old, new):
            if item == old:
                return new
            return item

        def replace_left(item, old, new):
            if item[: len(old)] == old:
                return new + item[len(old) :]
            return item

        classes = new_obj.attributes["class"].split()
        classes = map(lambda x: replace(x, "text-left", "text-start"), classes)
        classes = map(lambda x: replace(x, "text-right", "text-end"), classes)
        classes = map(lambda x: replace(x, "float-left", "float-start"), classes)
        classes = map(lambda x: replace(x, "float-right", "float-end"), classes)
        classes = map(lambda x: replace(x, "border-left", "border-start"), classes)
        classes = map(lambda x: replace(x, "border-right", "border-end"), classes)
        classes = map(lambda x: replace(x, "no-gutter", "g-0"), classes)
        classes = map(lambda x: replace(x, "text-monospace", "font-monospace"), classes)
        classes = map(lambda x: replace(x, "sr-only", "visually-hidden"), classes)
        classes = map(lambda x: replace_left(x, "left-", "start-"), classes)
        classes = map(lambda x: replace_left(x, "right-", "end-"), classes)
        classes = map(lambda x: replace_left(x, "ml-", "ms-"), classes)
        classes = map(lambda x: replace_left(x, "mr-", "me-"), classes)
        classes = map(lambda x: replace_left(x, "pl-", "ps-"), classes)
        classes = map(lambda x: replace_left(x, "pr-", "pe-"), classes)
        new_obj.config["attributes"]["class"] = " ".join(classes)


def x002_replace_card_deck(obj, new_obj):
    if obj.card_type == "card-deck":
        print("=> Detected bootstrap v4 card-deck which is not part of bootstrap5")
        print("   Replaced it with Card Layout, grid cards")
        new_obj.config["card_type"] = "row"
    if obj.card_type == "card-deck" or obj.card_type == "card-group":
        new_obj.plugin_type = "CardLayoutPlugin"
        new_obj.ui_item = "CardLayout"
    classes = obj.attributes.get("class", "").split()
    if "h-100" in classes:
        classes.remove("h-100")
        new_obj.config["card_full_height"] = True
        new_obj.config["attributes"]["class"] = " ".join(classes)
    if new_obj.config["card_alignment"] == "text-left":
        new_obj.config["card_alignment"] = "text-start"
    elif new_obj.config["card_alignment"] == "text-right":
        new_obj.config["card_alignment"] = "text-end"


def x003_card_context(obj, new_obj):
    if obj.card_outline:
        new_obj.config["card_outline"] = new_obj.config["background_context"]
        new_obj.config["background_context"] = ""
    else:
        new_obj.config["card_outline"] = ""


def a001_alignment(obj, new_obj, field):
    if field in new_obj.config and new_obj.config[field]:
        new_obj.config[field].replace("text-left", "start")
        new_obj.config[field].replace("text-center", "center")
        new_obj.config[field].replace("text-right", "end")


def m001_spacing_mixin(obj, new_obj, type):
    classes = new_obj.config["attributes"].get("class", "").split()
    if classes:
        for size, _ in list(settings.SPACER_SIZE_CHOICES) + (
            [("auto", "auto")] if type == "margin" else []
        ):
            if f"{type[0]}-{size}" in classes:
                classes.remove(f"{type[0]}-{size}")
                classes.append(f"{type[0]}x-{size}")
                classes.append(f"{type[0]}y-{size}")
            for side, _ in settings.SPACER_X_SIDES_CHOICES:
                if f"{type[0]}{side}-{size}" in classes and not new_obj.config.get(
                    f"{type}_x", None
                ):
                    new_obj.config[f"{type}_x"] = f"{type[0]}{side}-{size}"
                    new_obj.config[f"{type}_devices"] = None
                    classes.remove(f"{type[0]}{side}-{size}")
            for side, _ in settings.SPACER_Y_SIDES_CHOICES:
                if f"{type[0]}{side}-{size}" in classes and not new_obj.config.get(
                    f"{type}_y", None
                ):
                    new_obj.config[f"{type}_y"] = f"{type[0]}{side}-{size}"
                    new_obj.config[f"{type}_devices"] = None
                    classes.remove(f"{type[0]}{side}-{size}")
        if classes:
            new_obj.config["attributes"]["class"] = " ".join(classes)
        else:
            new_obj.config["attributes"].pop("class")


def m002_responsive_mixin(obj, new_obj):
    classes = new_obj.config["attributes"].get("class", "").split()
    if classes:
        display = (
            "block",
            "flex",
        )
        hidden = "none"

        visible = True
        hit = False
        responsive = []

        for device, _ in settings.DEVICE_CHOICES:
            stump = f"d-{device}-" if device != "xs" else "d-"
            if f"{stump}{hidden}" in classes and visible:
                visible = False
                hit = True
                classes.remove(f"{stump}{hidden}")
            for type in display:
                if f"{stump}{type}" in classes and not visible:
                    visible = True
                    hit = True
                    classes.remove(f"{stump}{type}")
            if visible:
                responsive.append(device)
        if hit:
            new_obj.config["responsive_visibility"] = responsive
            if classes:
                new_obj.config["attributes"]["class"] = " ".join(classes)
            else:
                new_obj.config["attributes"].pop("class")
        else:
            new_obj.config["responsive_visibility"] = None


def m003_background_mixin(obj, new_obj):
    classes = new_obj.config["attributes"].get("class", "").split()
    if classes:
        for context, _ in settings.COLOR_STYLE_CHOICES:
            if f"bg-{context}" in classes:
                new_obj.config["background_context"] = context
                classes.remove(f"bg-{context}")
        for cls, key in {
            "shadow-none": "none",
            "shadow-sm": "sm",
            "shadow": "reg",
            "shadow-lg": "lg",
        }.items():
            if cls in classes:
                new_obj.config["background_shadow"] = key
                classes.remove(cls)
        if classes:
            new_obj.config["attributes"]["class"] = " ".join(classes)
        else:
            new_obj.config["attributes"].pop("class")


def g001_col_text_alignment(obj, new_obj):
    if obj.column_type != "col":
        print(f"Warning: Break column detected - not supported (id={obj.id})")
    classes = new_obj.config["attributes"].get("class", "").split()
    if "text-left" in classes or "text-start" in classes:
        classes.remove("text-left")
        classes.remove("text-start")
        new_obj.config["text_alignment"] = "start"
    if "text-center" in classes:
        classes.remove("text-center")
        new_obj.config["text_alignment"] = "center"
    if "text-right" in classes or "text-end" in classes:
        classes.remove("text-right")
        classes.remove("text-end")
        new_obj.config["text_alignment"] = "end"
    if classes:
        new_obj.config["attributes"]["class"] = " ".join(classes)
    elif "class" in new_obj.config["attributes"]:
        new_obj.config["attributes"].pop("class")


def in_choices(choice, choices):
    for key, value in choices:
        if isinstance(value, (tuple, list)):
            if in_choices(choice, value):
                return True
        else:
            if key == choice:
                return True
    return False


def t001_template(obj, new_obj, bs4_setting, dcf_setting):
    if obj.template == "default":
        return
    BS4 = getattr(django_setting, bs4_setting, ())
    DCF = getattr(django_setting, dcf_setting, ())
    if not in_choices(obj.template, BS4):
        print(f"=> Template '{obj.template}' in {obj.plugin_type} (id: {obj.id})")
        print(f"   but not declared in {bs4_setting}")
        print(
            "   You will be able to edit the plugin but up saving the template will be changed."
        )
        if not DCF and bs4_setting != dcf_setting:
            print(f"   Remember to put {dcf_setting} in your settings.py")


data_migration = {
    "P001": p001_left_right_migration,
    "X002": x002_replace_card_deck,
    "X003": x003_card_context,
    "A001_quote": lambda x, y: a001_alignment(x, y, "quote_alignment"),
    "A001_figure": lambda x, y: a001_alignment(x, y, "figure_alignment"),
    "A001_picture": lambda x, y: a001_alignment(x, y, "alignment"),
    "A001_card": lambda x, y: a001_alignment(x, y, "card_alignment"),
    "G001": g001_col_text_alignment,
    "M001-m": lambda x, y: m001_spacing_mixin(x, y, "margin"),
    "M001-p": lambda x, y: m001_spacing_mixin(x, y, "padding"),
    "M002": m002_responsive_mixin,
    "M003": m003_background_mixin,
    "T001_PICTURE": lambda x, y: t001_template(
        x, y, "DJANGOCMS_PICTURE_TEMPLATES", "DJANGOCMS_PICTURE_TEMPLATES"
    ),
    "T001_LINK": lambda x, y: t001_template(
        x, y, "DJANGOCMS_LINK_TEMPLATES", "DJANGOCMS_FRONTEND_LINK_TEMPLATE_CHOICES"
    ),
    "T001_TABS": lambda x, y: t001_template(
        x, y, "DJANGOCMS_BOOTSTRAP4_TAB_TEMPLATES", "DJANGOCMS_FRONTEND_TAB_TEMPLATES"
    ),
    "T001_CAROUSEL": lambda x, y: t001_template(
        x,
        y,
        "DJANGOCMS_BOOTSTRAP4_CAROUSEL_TEMPLATES",
        "DJANGOCMS_FRONTEND_CAROUSEL_TEMPLATES",
    ),
    "T001_CAROUSEL_SLIDE": lambda x, y: t001_template(
        x,
        y,
        "DJANGOCMS_BOOTSTRAP4_CAROUSEL_TEMPLATES",
        "DJANGOCMS_FRONTEND_CAROUSEL_TEMPLATES",
    ),
}

plugin_prefix = "Bootstrap4"



from django import forms
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelForm

from djangocms_frontend import settings
from djangocms_frontend.common.background import BackgroundFormMixin
from djangocms_frontend.common.responsive import ResponsiveFormMixin
from djangocms_frontend.common.spacing import SpacingFormMixin
from djangocms_frontend.contrib import jumbotron
from djangocms_frontend.fields import (
    AttributesFormField,
    TagTypeFormField,
    TemplateChoiceMixin,
)
from djangocms_frontend.helpers import first_choice

mixin_factory = settings.get_forms(jumbotron)


class JumbotronForm(
    mixin_factory("Jumbotron"),
    ResponsiveFormMixin,
    SpacingFormMixin,
    BackgroundFormMixin,
    TemplateChoiceMixin,
    EntangledModelForm,
):
    """
    Components > "Jumbotron" Plugin
    https://getbootstrap.com/docs/5.0/components/jumbotron/
    """

    class Meta:
        entangled_fields = {
            "config": [
                "jumbotron_fluid",
                "template",
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    template = forms.ChoiceField(
        label=_("Template"),
        choices=settings.JUMBOTRON_TEMPLATE_CHOICES,
        initial=first_choice(settings.JUMBOTRON_TEMPLATE_CHOICES),
    )
    jumbotron_fluid = forms.BooleanField(
        label=_("Fluid"),
        initial=False,
        required=False,
        help_text=_(
            "Makes the jumbotron fill the full width of the container or window."
        ),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()



from cms.api import add_plugin
from cms.test_utils.testcases import CMSTestCase

from djangocms_frontend.contrib.carousel.cms_plugins import (
    CarouselPlugin,
    CarouselSlidePlugin,
)
from djangocms_frontend.contrib.carousel.forms import CarouselForm, CarouselSlideForm

from ..fixtures import TestFixture
from ..helpers import get_filer_image


class CarouselPluginTestCase(TestFixture, CMSTestCase):
    def setUp(self):
        super().setUp()
        self.image = get_filer_image()

    def tearDown(self):
        super().tearDown()
        self.image.delete()

    def test_carousel_plugin(self):
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=CarouselPlugin.__name__,
            language=self.language,
        )
        plugin.initialize_from_form(CarouselForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(
            ('<div class="slide carousel"' in response.content.decode("utf-8"))
            or ('<div class="carousel slide"' in response.content.decode("utf-8")),
            f'<div class="slide carousel" not found in {response.content.decode("utf-8")}',
        )
        self.assertContains(response, 'data-bs-interval="5000"')
        self.assertContains(response, 'data-bs-keyboard="true"')
        self.assertContains(response, 'data-bs-pause="hover"')
        self.assertContains(response, 'data-bs-ride="carousel"')
        self.assertContains(response, 'data-bs-wrap="true"')

    def test_carousel_slide_plugin(self):
        row = add_plugin(
            placeholder=self.placeholder,
            plugin_type=CarouselPlugin.__name__,
            language=self.language,
        )
        row.initialize_from_form(CarouselForm).save()
        plugin = add_plugin(
            target=row,
            placeholder=self.placeholder,
            plugin_type=CarouselSlidePlugin.__name__,
            language=self.language,
            config=dict(carousel_image=dict(pk=self.image.id, model="filer.Image")),
        )
        plugin.initialize_from_form(CarouselSlideForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(
            ('<div class="slide carousel"' in response.content.decode("utf-8"))
            or ('<div class="carousel slide"' in response.content.decode("utf-8")),
            f'<div class="slide carousel" not found in {response.content.decode("utf-8")}',
        )

        # testing aspect ratio variant
        # also testing multiply entries
        row = add_plugin(
            placeholder=self.placeholder,
            plugin_type=CarouselPlugin.__name__,
            language=self.language,
            config=dict(
                carousel_aspect_ratio="16x9",
            ),
        )
        row.initialize_from_form(CarouselForm).save()
        add_plugin(
            target=row,
            placeholder=self.placeholder,
            plugin_type=CarouselSlidePlugin.__name__,
            language=self.language,
            config=dict(carousel_image=dict(pk=self.image.id, model="filer.Image")),
        ).initialize_from_form(CarouselSlideForm).save()
        add_plugin(
            target=row,
            placeholder=self.placeholder,
            plugin_type=CarouselSlidePlugin.__name__,
            language=self.language,
            config=dict(carousel_image=dict(pk=self.image.id, model="filer.Image")),
        ).initialize_from_form(CarouselSlideForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(
            ('<div class="slide carousel"' in response.content.decode("utf-8"))
            or ('<div class="carousel slide"' in response.content.decode("utf-8")),
            f'<div class="slide carousel" not found in {response.content.decode("utf-8")}',
        )



from cms.api import add_plugin
from cms.test_utils.testcases import CMSTestCase

from djangocms_frontend.contrib.alert.cms_plugins import AlertPlugin
from djangocms_frontend.contrib.alert.forms import AlertForm

from ..fixtures import TestFixture


class AlertPluginTestCase(TestFixture, CMSTestCase):
    def test_plugin(self):
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=AlertPlugin.__name__,
            language=self.language,
        )
        plugin.initialize_from_form(AlertForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(
            (
                '<div class="alert alert-primary" role="alert">'
                in response.content.decode("utf-8")
            )
            or (
                '<div class="alert-primary alert" role="alert">'
                in response.content.decode("utf-8")
            ),
            f'alert-primary not found in {response.content.decode("utf-8")}',
        )



from djangocms_frontend import settings

foundation_sizes = dict(xs="small", md="medium", xl="large")


class GridContainerRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes("grid-container")
        if instance.container_type == "container-fluid":
            instance.add_classes("fluid")
        elif instance.container_type != "container":
            instance.add_classss("full")
        return super().render(context, instance, placeholder)


def get_row_cols_grid_values(instance):
    classes = []
    for device in settings.DEVICE_SIZES:
        size = getattr(instance, f"row_cols_{device}", None)
        if isinstance(size, int):
            classes.append(f"{foundation_sizes.get(device, device)}-up-{int(size)}")
    return classes


class GridRowRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes("grid-x", get_row_cols_grid_values(instance))
        if not instance.gutters:
            instance.add_classes("grid-padding-x", "grid-padding-y")
        return super().render(context, instance, placeholder)


def get_grid_values(self):
    classes = []
    for device in settings.DEVICE_SIZES:
        for element in ("col", "order", "offset"):
            size = getattr(self, f"{device}_{element}", None)
            if isinstance(size, int):
                if element == "col":
                    if size == 0:
                        classes.append(f"{foundation_sizes.get(device, device)}-auto")
                    else:
                        classes.append(f"{foundation_sizes.get(device, device)}-{size}")
                else:
                    classes.append(
                        f"{foundation_sizes.get(device, device)}-{element}-{size}"
                    )

    return classes


class GridColumnRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes("cell", get_grid_values(instance))
        return super().render(context, instance, placeholder)



class SpacingRenderMixin:
    def render(self, context, instance, placeholder):
        if not instance.space_device or instance.space_device == "xs":
            instance.add_classes(
                f"{instance.space_property}{instance.space_sides}-{instance.space_size}"
            )
        else:
            instance.add_classes(
                f"{instance.space_property}{instance.space_sides}-{ instance.space_device }-{instance.space_size}"
            )
        return super().render(context, instance, placeholder)


class EditorNoteRenderMixin:
    render_template = "djangocms_frontend/bootstrap5/editor_note.html"


class HeadingRenderMixin:
    def render(self, context, instance, placeholder):
        heading_context = instance.config.get("heading_context", None)
        if heading_context:
            instance.add_classes(f"text-{heading_context}")
        heading_alignment = instance.config.get("heading_alignment", None)
        if heading_alignment:
            instance.add_classes(f"text-{heading_alignment}")
        return super().render(context, instance, placeholder)



from cms.api import add_plugin
from cms.test_utils.testcases import CMSTestCase

from djangocms_frontend.contrib.content.cms_plugins import (
    BlockquotePlugin,
    CodePlugin,
    FigurePlugin,
)
from djangocms_frontend.contrib.content.forms import BlockquoteForm, FigureForm

from ..fixtures import TestFixture


class ContentPluginTestCase(TestFixture, CMSTestCase):
    def test_code_plugin(self):
        add_plugin(
            placeholder=self.placeholder,
            plugin_type=CodePlugin.__name__,
            language=self.language,
            config=dict(
                code_content="<p>hello world</p>",
            ),
        )
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "&lt;p&gt;hello world&lt;/p&gt;")

    def test_blockquote_plugin(self):
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=BlockquotePlugin.__name__,
            language=self.language,
            config=dict(
                quote_content="hello world",
            ),
        )
        plugin.initialize_from_form(BlockquoteForm)
        plugin.save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<blockquote class="blockquote">')

        # test quote_alignment
        add_plugin(
            placeholder=self.placeholder,
            plugin_type=BlockquotePlugin.__name__,
            language=self.language,
            config=dict(
                quote_content="hello world",
                quote_alignment="",
            ),
        )
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<blockquote class="blockquote">')

    def test_figure_plugin(self):
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=FigurePlugin.__name__,
            language=self.language,
            config=dict(
                figure_caption="hello world",
            ),
        )
        plugin.initialize_from_form(FigureForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<figcaption class="figure-caption')



from django import forms
from django.conf import settings as django_settings
from django.db.models.fields.related import ManyToOneRel
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelForm
from filer.fields.image import AdminImageFormField, FilerImageField
from filer.models import Image, ThumbnailOption

from djangocms_frontend import settings

from ...common.responsive import ResponsiveFormMixin
from ...common.spacing import MarginFormMixin
from ...fields import AttributesFormField, TagTypeFormField, TemplateChoiceMixin
from ...helpers import first_choice
from ...models import FrontendUIItem
from ..link.forms import AbstractLinkForm


def get_alignment():
    """add setting for image alignment, renders a class or inline styles depending on your template setup"""
    alignment = getattr(
        settings,
        "DJANGOCMS_PICTURE_ALIGN",
        (
            ("start", _("Align left")),
            ("end", _("Align right")),
            ("center", _("Align center")),
        ),
    )
    return alignment


def get_templates():
    """Add additional choices through the ``settings.py``."""
    choices = getattr(
        django_settings,
        "DJANGOCMS_PICTURE_TEMPLATES",
        [
            ("default", _("Default")),
        ],
    )
    return choices


# required for backwards compability
PICTURE_ALIGNMENT = get_alignment()


LINK_TARGET = (
    ("_blank", _("Open in new window")),
    ("_self", _("Open in same window")),
    ("_parent", _("Delegate to parent")),
    ("_top", _("Delegate to top")),
)

RESPONSIVE_IMAGE_CHOICES = (
    ("inherit", _("Let settings.DJANGOCMS_PICTURE_RESPONSIVE_IMAGES decide")),
    ("yes", _("Yes")),
    ("no", _("No")),
)


class ImageForm(
    TemplateChoiceMixin,
    AbstractLinkForm,
    ResponsiveFormMixin,
    MarginFormMixin,
    EntangledModelForm,
):
    """
    Content > "Image" Plugin
    https://getbootstrap.com/docs/5.0/content/images/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "template",
                "picture",
                "external_picture",
                "lazy_loading",
                "width",
                "height",
                "alignment",
                "link_attributes",
                "use_automatic_scaling",
                "use_crop",
                "use_no_cropping",
                "use_upscale",
                "use_responsive_image",
                "thumbnail_options",
                "picture_fluid",
                "picture_rounded",
                "picture_thumbnail",
                "attributes",
            ]
        }

    link_is_optional = True

    template = forms.ChoiceField(
        label=_("Template"),
        choices=get_templates(),
        initial=first_choice(get_templates()),
    )
    picture = AdminImageFormField(
        rel=ManyToOneRel(FilerImageField, Image, "id"),
        queryset=Image.objects.all(),
        to_field_name="id",
        label=_("Image"),
        required=False,
    )
    external_picture = forms.URLField(
        label=_("External image"),
        required=False,
        help_text=_(
            "If provided, overrides the embedded image. "
            "Certain options such as cropping are not applicable to external images."
        ),
    )
    lazy_loading = forms.BooleanField(
        label=_("Load lazily"),
        required=False,
        help_text=_(
            "Use for images below the fold. This will load images only if user scrolls them into view. "
        ),
    )

    width = forms.IntegerField(
        label=_("Width"),
        required=False,
        min_value=1,
        help_text=_(
            "The image width as number in pixels. " 'Example: "720" and not "720px".'
        ),
    )
    height = forms.IntegerField(
        label=_("Height"),
        required=False,
        min_value=1,
        help_text=_(
            "The image height as number in pixels. " 'Example: "720" and not "720px".'
        ),
    )
    alignment = forms.ChoiceField(
        label=_("Alignment"),
        choices=get_alignment(),
        required=False,
        help_text=_("Aligns the image according to the selected option."),
    )
    link_attributes = AttributesFormField(
        label=_("Link attributes"),
        help_text=_("Attributes apply to the <b>link</b>."),
    )

    # cropping models
    # active per default
    use_automatic_scaling = forms.BooleanField(
        label=_("Automatic scaling"),
        required=False,
        help_text=_(
            "Uses the placeholder dimensions to automatically calculate the size."
        ),
    )
    # ignores all other cropping options
    # throws validation error if other cropping options are selected
    use_no_cropping = forms.BooleanField(
        label=_("Use original image"),
        required=False,
        help_text=_("Outputs the raw image without cropping."),
    )
    # upscale and crop work together
    # throws validation error if other cropping options are selected
    use_crop = forms.BooleanField(
        label=_("Crop image"),
        required=False,
        help_text=_(
            "Crops the image according to the thumbnail settings provided in the template."
        ),
    )
    use_upscale = forms.BooleanField(
        label=_("Upscale image"),
        required=False,
        help_text=_(
            "Upscales the image to the size of the thumbnail settings in the template."
        ),
    )
    use_responsive_image = forms.ChoiceField(
        label=_("Use responsive image"),
        choices=RESPONSIVE_IMAGE_CHOICES,
        initial=first_choice(RESPONSIVE_IMAGE_CHOICES),
        help_text=_(
            "Uses responsive image technique to choose better image to display based upon screen viewport. "
            "This configuration only applies to uploaded images (external pictures will not be affected). "
        ),
    )
    # overrides all other options
    # throws validation error if other cropping options are selected
    thumbnail_options = forms.ModelChoiceField(
        queryset=ThumbnailOption.objects.all(),
        to_field_name="id",
        label=_("Thumbnail options"),
        required=False,
        help_text=_(
            "Overrides width, height, and crop; scales up to the provided preset dimensions."
        ),
    )
    picture_fluid = forms.BooleanField(
        label=_("Responsive"),
        required=False,
        initial=True,
        help_text=_("Adds the .img-fluid class to make the image responsive."),
    )
    picture_rounded = forms.BooleanField(
        label=_("Rounded"),
        required=False,
        initial=False,
        help_text=_("Adds the .rounded class for round corners."),
    )
    picture_thumbnail = forms.BooleanField(
        label=_("Thumbnail"),
        required=False,
        initial=False,
        help_text=_("Adds the .img-thumbnail class."),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()

    def clean(self):
        super().clean()
        data = self.cleaned_data
        # there can be only one link type
        if (
            sum(
                (
                    bool(data.get("external_link", False)),
                    bool(data.get("internal_link", False)),
                    bool(data.get("file_link", False)),
                )
            )
            > 1
        ):
            raise forms.ValidationError(
                _(
                    "You have given more than one external, internal, or file link target. "
                    "Only one option is allowed."
                )
            )

        # you shall only set one image kind
        if not data.get("picture", False) and not data.get("external_picture", False):
            raise forms.ValidationError(
                _(
                    "You need to add either an image, "
                    "or a URL linking to an external image."
                )
            )

        # certain cropping options do not work together, the following
        # list defines the disallowed options used in the ``clean`` method
        invalid_option_pairs = [
            ("use_automatic_scaling", "use_no_cropping"),
            ("use_automatic_scaling", "thumbnail_options"),
            ("use_no_cropping", "use_crop"),
            ("use_no_cropping", "use_upscale"),
            ("use_no_cropping", "thumbnail_options"),
            ("thumbnail_options", "use_crop"),
            ("thumbnail_options", "use_upscale"),
        ]
        # invalid_option_pairs
        invalid_option_pair = None

        for pair in invalid_option_pairs:
            if data.get(pair[0], False) and data.get(pair[1], False):
                invalid_option_pair = pair
                break

        if invalid_option_pair:
            message = _(
                "Invalid cropping settings. "
                'You cannot combine "{field_a}" with "{field_b}".'
            )
            message = message.format(
                field_a=self.fields[invalid_option_pair[0]].label,
                field_b=self.fields[invalid_option_pair[0]].label,
            )
            raise forms.ValidationError(message)



from importlib import import_module

from djangocms_frontend import settings

try:
    module = import_module(f"..{settings.framework}.sizing", __name__)
    SizingFormMixin = module.SizingFormMixin
    SizingMixin = module.SizingMixin
except ModuleNotFoundError:

    class SizingMixin:
        pass

    class SizingFormMixin:
        pass



from django.test import TestCase

from djangocms_frontend.contrib.carousel.forms import CarouselForm, CarouselSlideForm
from djangocms_frontend.contrib.carousel.models import Carousel, CarouselSlide

from ..helpers import get_filer_image


class CarouselModelTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.image = get_filer_image()

    def tearDown(self):
        super().tearDown()
        self.image.delete()

    def test_carousel_instance(self):
        instance = Carousel.objects.create()
        instance.initialize_from_form(CarouselForm).save()
        self.assertEqual(str(instance), "Carousel (1)")
        self.assertEqual(
            instance.get_short_description(),
            "(default) Interval: 5000, Controls: True, Indicators: True, "
            "Keyboard: True, Pause: hover, Ride: True, Wrap: True",
        )

    def test_carousel_slide_instance(self):
        instance = CarouselSlide.objects.create()
        instance.initialize_from_form(CarouselSlideForm).save()
        self.assertEqual(str(instance), "CarouselSlide (1)")
        self.assertEqual(instance.get_short_description(), "")
        # test carousel content strings
        instance.config["carousel_image"] = dict(pk=self.image.id, model="filer.Image")
        instance.config["carousel_content"] = "hello world"
        self.assertEqual(
            instance.get_short_description(),
            "test_file.jpg (hello world)",
        )
        instance.config["carousel_content"] = "hello world" + 100 * "#"
        self.assertEqual(
            instance.get_short_description(),
            "test_file.jpg (hello world" + "#" * 89 + "...)",
        )
        self.assertEqual(instance.get_link(), "")
        # # test image output options
        # instance.config["carousel_content"] = None
        # instance.config["carousel_image"] = dict(pk=get_filer_image(name="image").id, model="filer.Image")
        # self.assertEqual(instance.get_short_description(), "image")
        # instance.config["carousel_image"] = dict(
        #     pk=get_filer_image(name="image", original_filename=False).id, model="filer.Image"
        # )
        # self.assertEqual(instance.get_short_description(), "Image")



from django.test import TestCase

from djangocms_frontend.contrib.utilities.models import (
    Heading,
    Spacing,
    TableOfContents,
)


class B5UtilitiesModelTestCase(TestCase):
    def test_instance(self):
        instance = Spacing.objects.create(
            config={
                "space_device": "",
                "space_property": "m",
                "space_size": 0,
                "space_sides": "",
            }
        )
        self.assertEqual(str(instance), "Spacing (1)")
        self.assertEqual(instance.get_short_description(), "(.m-0)")

        instance.config["space_device"] = "md"
        self.assertEqual(instance.get_short_description(), "(.m-md-0)")

        instance = Heading.objects.create(
            config=dict(
                heading_level="h2",
                heading="Welcome to Django-CMS!",
            )
        )
        self.assertEqual(str(instance), "Heading (2)")
        self.assertEqual(instance.get_short_description(), "(Welcome to Django-CMS!)")

        instance = TableOfContents.objects.create()
        self.assertEqual(str(instance), "TableOfContents (3)")
        self.assertEqual(instance.get_short_description(), "(3)")



from django import forms
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelForm

from djangocms_frontend import settings

from ...common.responsive import ResponsiveFormMixin
from ...common.spacing import MarginFormMixin, PaddingFormMixin
from ...fields import (
    AttributesFormField,
    ButtonGroup,
    ColoredButtonGroup,
    TagTypeFormField,
)
from ...models import FrontendUIItem
from .constants import LISTGROUP_STATE_CHOICES


class ListGroupForm(MarginFormMixin, ResponsiveFormMixin, EntangledModelForm):
    """
    Components > "List Group" Plugin
    https://getbootstrap.com/docs/5.0/components/list-group/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "list_group_flush",
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    list_group_flush = forms.BooleanField(
        label=_("List group flush"),
        initial=False,
        required=False,
        help_text=_("Create lists of content in a card with a flush list group."),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()


class ListGroupItemForm(PaddingFormMixin, ResponsiveFormMixin, EntangledModelForm):
    """
    Components > "List Group Item" Plugin
    https://getbootstrap.com/docs/5.0/components/list-group/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "simple_content",
                "list_context",
                "list_state",
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    simple_content = forms.CharField(
        label=_("One line content"),
        required=False,
        help_text=_(
            "List item text. Is only show if this list item has no child plugins."
        ),
    )
    list_context = forms.ChoiceField(
        label=_("Context"),
        choices=settings.EMPTY_CHOICE + settings.COLOR_STYLE_CHOICES,
        initial=settings.EMPTY_CHOICE[0][0],
        required=False,
        widget=ColoredButtonGroup(),
    )
    list_state = forms.ChoiceField(
        label=_("State"),
        choices=settings.EMPTY_CHOICE + LISTGROUP_STATE_CHOICES,
        initial=settings.EMPTY_CHOICE[0][0],
        required=False,
        widget=ButtonGroup(attrs=dict(property="list_state")),
    )
    attributes = AttributesFormField()
    tag_type = TagTypeFormField()



from djangocms_frontend import settings


def get_row_cols_grid_values(instance):
    classes = []
    for device in settings.DEVICE_SIZES:
        size = getattr(instance, f"row_cols_{device}", None)
        if isinstance(size, int):
            if device == "xs":
                classes.append(f"row-cols-{int(size)}")
            else:
                classes.append(f"row-cols-{device}-{int(size)}")
    return classes


class GridContainerRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes(instance.container_type)
        return super().render(context, instance, placeholder)


class GridRowRenderMixin:
    render_template = "djangocms_frontend/bootstrap5/grid_row.html"

    def render(self, context, instance, placeholder):
        instance.add_classes(
            "row",
            instance.vertical_alignment,
            instance.horizontal_alignment,
        )
        if instance.parent and instance.parent.plugin_type == "CardPlugin":
            instance.add_classes("g-0")  # no gutters if inside card
        if instance.config.get("gutters", ""):
            instance.add_classes(f"g-{instance.gutters}")
        instance.add_classes(get_row_cols_grid_values(instance))
        return super().render(context, instance, placeholder)


def get_grid_values(self):
    classes = []
    for device in settings.DEVICE_SIZES:
        for element in ("col", "order", "offset", "ms", "me"):
            size = getattr(self, f"{device}_{element}", None)
            if isinstance(size, int) and (
                element == "col" or element == "order" or element == "offset"
            ):
                if size == 0 and element == "col":  # 0 represents auto
                    size = "auto"
                if device == "xs":
                    classes.append(f"{element}-{size}")
                else:
                    classes.append(f"{element}-{device}-{size}")
            elif size:
                if device == "xs":
                    classes.append("{}-{}".format(element, "auto"))
                else:
                    classes.append("{}-{}-{}".format(element, device, "auto"))

    return classes


class GridColumnRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes(
            f"col text-{instance.text_alignment}"
            if instance.config.get("text_alignment", None)
            else "col"
        )
        instance.add_classes(instance.column_alignment)
        instance.add_classes(get_grid_values(instance))
        return super().render(context, instance, placeholder)



class BlockquoteRenderMixin:
    render_template = "djangocms_frontend/bootstrap5/blockquote.html"

    def render(self, context, instance, placeholder):
        if not instance.config.get("quote_source", None):
            instance.add_classes("blockquote")
        if instance.quote_alignment:
            instance.add_classes(f"text-{instance.quote_alignment}")
        return super().render(context, instance, placeholder)


class FigureRenderMixin:
    render_template = "djangocms_frontend/bootstrap5/figure.html"

    def render(self, context, instance, placeholder):
        instance.add_classes("figure")
        if instance.figure_alignment:
            instance.add_classes(f"text-{instance.figure_alignment}")
        return super().render(context, instance, placeholder)


class CodeRenderMixin:
    render_template = "djangocms_frontend/bootstrap5/code.html"



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from ... import settings
from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...common.background import BackgroundMixin
from ...common.responsive import ResponsiveMixin
from ...common.spacing import MarginMixin, PaddingMixin
from ...helpers import add_plugin
from .. import card
from . import forms, models

mixin_factory = settings.get_renderer(card)


@plugin_pool.register_plugin
class CardLayoutPlugin(mixin_factory("CardLayout"), AttributesMixin, CMSUIPlugin):
    """
    Components > "Card" Plugin
    https://getbootstrap.com/docs/5.0/components/card/
    """

    name = _("Card layout")
    module = _("Frontend")
    model = models.CardLayout
    form = forms.CardLayoutForm
    change_form_template = "djangocms_frontend/admin/card_layout.html"
    allow_children = True
    child_classes = [
        "CardPlugin",
    ]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    (
                        "card_type",
                        "create",
                    ),
                )
            },
        ),
        (
            _("Responsive settings"),
            {
                "fields": ([f"row_cols_{size}" for size in settings.DEVICE_SIZES],),
            },
        ),
    ]

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        data = form.cleaned_data
        for pos in range(data["create"] if data["create"] is not None else 0):
            add_plugin(
                obj.placeholder,
                models.Card(
                    parent=obj,
                    placeholder=obj.placeholder,
                    position=obj.position + 1 + pos,
                    language=obj.language,
                    plugin_type=CardPlugin.__name__,
                    ui_item=models.Card.__class__.__name__,
                ).initialize_from_form(forms.CardForm),
            )


@plugin_pool.register_plugin
class CardPlugin(
    mixin_factory("Card"),
    AttributesMixin,
    ResponsiveMixin,
    MarginMixin,
    BackgroundMixin,
    CMSUIPlugin,
):
    """
    Components > "Card" Plugin
    https://getbootstrap.com/docs/5.0/components/card/
    """

    name = _("Card")
    module = _("Frontend")
    model = models.Card
    form = forms.CardForm
    change_form_template = "djangocms_frontend/admin/card.html"
    allow_children = True
    child_classes = [
        "CardInnerPlugin",
        "ListGroupPlugin",
        "ImagePlugin",
        "GridRowPlugin",
    ]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "card_alignment",
                    (
                        "card_text_color",
                        "card_outline",
                    ),
                    "card_full_height",
                )
            },
        ),
    ]

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if not change:
            add_plugin(
                obj.placeholder,
                models.CardInner(
                    parent=obj,
                    position=obj.position + 1,
                    placeholder=obj.placeholder,
                    language=obj.language,
                    plugin_type=CardInnerPlugin.__name__,
                    ui_item=models.CardInner.__class__.__name__,
                    config=dict(inner_type="card-body"),
                ),
            )


@plugin_pool.register_plugin
class CardInnerPlugin(
    mixin_factory("CardInner"),
    AttributesMixin,
    ResponsiveMixin,
    PaddingMixin,
    BackgroundMixin,
    CMSUIPlugin,
):
    """
    Components > "Card - Inner" Plugin (Header, Footer, Body)
    https://getbootstrap.com/docs/5.0/components/card/
    """

    name = _("Card inner")
    module = _("Frontend")
    model = models.CardInner
    form = forms.CardInnerForm
    change_form_template = "djangocms_frontend/admin/card.html"
    allow_children = True
    parent_classes = [
        "CardPlugin",
        "CollapseTriggerPlugin",
        "CollapseContainerPlugin",
        "GridColumnPlugin",
    ]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "inner_type",
                    "text_alignment",
                )
            },
        ),
    ]



from django import template
from django.templatetags.static import static

from djangocms_frontend.contrib.icon.conf import ICON_LIBRARIES

register = template.Library()


@register.inclusion_tag("djangocms_frontend/icon/add_css.html", takes_context=True)
def add_css_for_icon(context, icon):
    if icon.get("library", "") in ICON_LIBRARIES:
        css_link = ICON_LIBRARIES[icon.get("library")][1]
        if css_link:
            if "/" not in css_link:  # static link?
                css_link = static(
                    f"djangocms_frontend/icon/vendor/assets/stylesheets/{css_link}"
                )
            context["icon_css"] = css_link
    return context



from cms.api import add_plugin
from cms.test_utils.testcases import CMSTestCase
from django.http import HttpRequest

from djangocms_frontend import settings
from djangocms_frontend.contrib.link.cms_plugins import LinkPlugin
from djangocms_frontend.contrib.link.forms import LinkForm, SmartLinkField
from djangocms_frontend.contrib.link.helpers import get_choices

from ..fixtures import DJANGO_CMS4, TestFixture


class LinkPluginTestCase(TestFixture, CMSTestCase):
    def test_plugin(self):
        add_plugin(
            placeholder=self.placeholder,
            plugin_type=LinkPlugin.__name__,
            language=self.language,
            config=dict(
                external_link="https://www.divio.com",
            ),
        ).initialize_from_form(LinkForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'href="https://www.divio.com"')

        # add more options
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=LinkPlugin.__name__,
            language=self.language,
            config=dict(
                external_link="https://www.divio.com",
                link_context="primary",
                link_size="btn-sm",
                link_block=True,
                name="django CMS rocks!",
            ),
        )
        plugin.initialize_from_form(LinkForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "link-primary")
        self.assertContains(response, "btn-sm")
        self.assertContains(response, "d-block")

        # alternate version for link_type
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=LinkPlugin.__name__,
            language=self.language,
            config=dict(
                internal_link=dict(model="cms.page", pk=self.page.id),
                link_context="primary",
                link_type="btn",
                name="django CMS rocks!",
            ),
        )
        plugin.initialize_from_form(LinkForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "btn-primary")
        self.assertContains(response, 'href="/content/"')

        # alternate version broken link
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=LinkPlugin.__name__,
            language=self.language,
            config=dict(
                internal_link=dict(model="cms.page", pk=-3141),
                link_context="primary",
                link_type="btn",
                link_outline=True,
                name="django CMS rocks!",
            ),
        )
        plugin.initialize_from_form(LinkForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "\ndjango CMS rocks")

        # alternate version using link_outline
        plugin = add_plugin(
            placeholder=self.placeholder,
            plugin_type=LinkPlugin.__name__,
            language=self.language,
            config=dict(
                external_link="https://www.divio.com",
                link_context="primary",
                link_type="btn",
                link_outline=True,
            ),
        )
        plugin.initialize_from_form(LinkForm).save()
        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(
            ('class="btn btn-outline-primary"' in response.content.decode("utf-8"))
            or ('class="btn-outline-primary btn"' in response.content.decode("utf-8")),
            f'Cound not find class="btn btn-outline-primary" in {response.content.decode("utf-8")}',
        )

    def test_smart_link_field(self):
        slf = SmartLinkField()
        choices = get_choices(None)
        self.assertEqual("example.com", choices[0][0])  # Site name
        self.assertIn(("2-1", "home"), choices[0][1])

        cleaned = slf.clean("2-1")
        self.assertEqual(dict(model="cms.page", pk=1), cleaned)

        self.assertEqual(slf.prepare_value("blabla"), "")
        self.assertEqual(slf.prepare_value(dict(model="cms.page", pk=1)), "2-1")
        self.assertEqual(slf.prepare_value(self.home), "2-1")

    def test_link_form(self):
        request = HttpRequest()
        request.POST = {
            "template": settings.LINK_TEMPLATE_CHOICES[0][0],
            "link_type": "link",
        }
        form = LinkForm(request.POST)
        self.assertFalse(form.is_valid())
        request.POST.update(
            {
                "name": "One of the world's most advanced open source CMS",
                "anchor": "allowed-here",
                "margin_devices": ["xs"],
                "padding_devices": ["xs"],
                **(
                    dict(
                        url_grouper=str(
                            self.create_url(manual_url="https://www.django-cms.org/").id
                        )
                    )
                    if DJANGO_CMS4
                    else dict(external_link="https://www.django-cms.org/")
                ),
            }
        )
        form = LinkForm(request.POST)
        self.assertTrue(form.is_valid())
        if DJANGO_CMS4:
            self.delete_urls()
        else:
            request.POST.update({"mailto": "none@nowhere.com"})
            form = LinkForm(request.POST)
            self.assertFalse(form.is_valid())  # Two targets
            request.POST["external_link"] = None
            form = LinkForm(request.POST)
            self.assertFalse(form.is_valid())  # no anchor for mail



from entangled.forms import EntangledModelForm

from djangocms_frontend.fields import AttributesFormField, TagTypeFormField

from ...common.responsive import ResponsiveFormMixin
from ...models import FrontendUIItem


class MediaForm(ResponsiveFormMixin, EntangledModelForm):
    """
    Layout > "Media" Plugin
    http://getbootstrap.com/docs/4.0/layout/media-object/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    attributes = AttributesFormField()
    tag_type = TagTypeFormField()


class MediaBodyForm(EntangledModelForm):
    """
    Layout > "Media body" Plugin
    http://getbootstrap.com/docs/4.0/layout/media-object/
    """

    class Meta:
        model = FrontendUIItem
        entangled_fields = {
            "config": [
                "attributes",
            ]
        }
        untangled_fields = ("tag_type",)

    attributes = AttributesFormField()
    tag_type = TagTypeFormField()



from django.apps import apps
from django.contrib.auth import get_user_model
from django.contrib.auth.management import _get_all_permissions
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType

from .base import SubcommandsCommand

User = get_user_model()


class SyncPermissions(SubcommandsCommand):
    help = "Sets the permissions for all plugins to those of djangocms_frontend.models.FrontendUIItem"
    command_name = "sync_permissions"

    def add_arguments(self, parser):
        parser.add_argument("scope", choices=["users", "groups"])

    def handle(self, *args, **options):
        if options["interactive"]:
            self.stdout.write(
                """This command changes permissions for djangocms-frontend plugins.
Changes cannot be undone. Are you sure you want to proceed?
"""
            )
            ok_to_delete = input("Type 'yes' to continue, or 'no' to cancel: ")
        else:
            ok_to_delete = "yes"

        if ok_to_delete != "yes":
            self.stdout.write(self.style.ERROR("Aborted."))
            return

        dcf_models = [
            m
            for m in apps.get_models()
            if m.__module__.startswith("djangocms_frontend.")
        ]
        reference_model = [m for m in dcf_models if m.__name__ == "FrontendUIItem"][0]
        dcf_models.remove(reference_model)
        ctype = ContentType.objects.get(
            app_label=reference_model._meta.app_label,
            model=reference_model._meta.object_name.lower(),
        )
        permission_set = Permission.objects.filter(content_type=ctype)
        if "users" in options["scope"]:
            for user in User.objects.all():
                self.stdout.write(self.style.SUCCESS(f"Snycing {user}"))
                reduced = permission_set.filter(user=user)
                reduced = [perm.codename.split("_")[0] for perm in reduced]
                if reduced:
                    self.stdout.write(
                        self.style.NOTICE(f"... has {', '.join(reduced)}")
                    )
                for model in dcf_models:
                    perms = _get_all_permissions(model._meta)
                    for perm, name in perms:
                        ok = perm.split("_")[0] in reduced
                        ctype = ContentType.objects.get(
                            app_label=model._meta.app_label,
                            model=model._meta.object_name.lower(),
                        )
                        new_perm, created = Permission.objects.get_or_create(
                            codename=perm,
                            content_type=ctype,
                            defaults=dict(name=name),
                        )
                        if created:
                            self.stdout.write(
                                self.style.WARNING(f"Created permission {new_perm}")
                            )
                        if ok:
                            if not user.user_permissions.filter(
                                codename=new_perm.codename, content_type=ctype
                            ):
                                self.stdout.write(
                                    self.style.NOTICE(f"Added {new_perm} for {user}")
                                )
                                user.user_permissions.add(new_perm)
                        else:
                            if user.user_permissions.filter(
                                codename=new_perm.codename, content_type=ctype
                            ):
                                self.stdout.write(
                                    self.style.WARNING(f"Removed {new_perm} for {user}")
                                )
                                user.user_permissions.remove(new_perm)
                user.save()
        if "groups" in options["scope"]:
            for group in Group.objects.all():
                self.stdout.write(self.style.SUCCESS(f"Snycing {group}"))
                reduced = permission_set.filter(group=group)
                reduced = [perm.codename.split("_")[0] for perm in reduced]
                if reduced:
                    self.stdout.write(
                        self.style.NOTICE(f"... has {', '.join(reduced)}")
                    )
                for model in dcf_models:
                    perms = _get_all_permissions(model._meta)
                    for perm, name in perms:
                        ok = perm.split("_")[0] in reduced
                        ctype = ContentType.objects.get(
                            app_label=model._meta.app_label,
                            model=model._meta.object_name.lower(),
                        )
                        new_perm, created = Permission.objects.get_or_create(
                            codename=perm,
                            content_type=ctype,
                            defaults=dict(name=name),
                        )
                        if ok:
                            if not group.permissions.filter(
                                codename=new_perm.codename, content_type=ctype
                            ):
                                self.stdout.write(
                                    self.style.NOTICE(f"Added {new_perm} for {group}")
                                )
                                group.permissions.add(new_perm)
                        else:
                            if group.permissions.filter(
                                codename=new_perm.codename, content_type=ctype
                            ):
                                self.stdout.write(
                                    self.style.WARNING(
                                        f"Removed {new_perm} for {group}"
                                    )
                                )
                                group.permissions.remove(new_perm)
                group.save()

        self.stdout.write(self.style.SUCCESS("Finished syncing permissions"))



from djangocms_frontend.frameworks.foundation6 import convert_context


class AlertRenderMixin:
    render_template = "djangocms_frontend/foundation6/alert.html"

    def render(self, context, instance, placeholder):
        instance.add_classes("callout", convert_context(instance.alert_context))
        if instance.alert_dismissible:
            instance.add_attribute("data-closable")
        return super().render(context, instance, placeholder)



class CardRenderMixin:
    def render(self, context, instance, placeholder):
        instance.add_classes("card")
        if instance.card_context and instance.card_outline:
            instance.add_classes(f"border-{instance.card_context}")
        elif instance.card_context:
            instance.add_classes(f"bg-{instance.card_context}")
        if instance.card_alignment:
            instance.add_classes(instance.card_alignment)
        if instance.card_text_color:
            instance.add_classes(f"text-{instance.card_text_color}")
        if getattr(instance, "card_full_height", False):
            instance.add_classes("h-100")
        if instance.parent and instance.parent.plugin_type == "CardLayoutPlugin":
            if instance.parent.get_plugin_instance()[0].card_type == "row":
                instance.add_classes("h-100")
        return super().render(context, instance, placeholder)


class CardInnerRenderMixin:
    def render(self, context, instance, placeholder):
        if instance.inner_type == "card-body":
            instance.add_classes("card-section")
        elif instance.inner_type in ("card-header", "card-footer"):
            instance.add_classes("card-divider")
        # if instance.card_context and instance.card_outline:
        #     instance.add_classes(f"border-{instance.card_context}")
        # elif instance.card_context:
        #     instance.add_classes(f"bg-{instance.card_context}")
        if getattr(instance, "card_alignment", ""):
            instance.add_classes(instance.card_alignment)
        # if instance.card_text_color:
        #     instance.add_classes(f"text-{instance.card_text_color}")
        # if getattr(instance, "card_full_height", False):
        #     instance.add_classes("h-100")
        if instance.parent and instance.parent.plugin_type == "CardLayoutPlugin":
            if instance.parent.get_plugin_instance()[0].card_type == "row":
                instance.add_classes("h-100")
        return super().render(context, instance, placeholder)



from django.contrib.sites.models import Site
from django.db import models
from django.utils.translation import gettext as _

from djangocms_frontend.helpers import get_related_object

# 'link' type is added manually as it is only required for this plugin
from djangocms_frontend.models import FrontendUIItem
from djangocms_frontend.settings import COLOR_STYLE_CHOICES

COLOR_STYLE_CHOICES = (("link", _("Link")),) + COLOR_STYLE_CHOICES


class GetLinkMixin:
    def get_link(self):
        if getattr(self, "url_grouper", None):
            url_grouper = get_related_object(self.config, "url_grouper")
            if not url_grouper:
                return ""
            url = url_grouper.get_content(show_draft_content=True)
            # simulate the call to the unauthorized CMSPlugin.page property
            cms_page = self.placeholder.page if self.placeholder_id else None

            # first, we check if the placeholder the plugin is attached to
            # has a page. Thus the check "is not None":
            if cms_page is not None:
                if getattr(cms_page, "node", None):
                    cms_page_site_id = getattr(cms_page.node, "site_id", None)
                else:
                    cms_page_site_id = getattr(cms_page, "site_id", None)
            # a plugin might not be attached to a page and thus has no site
            # associated with it. This also applies to plugins inside
            # static placeholders
            else:
                cms_page_site_id = None
            return url.get_url(cms_page_site_id) or ""

        if getattr(self, "internal_link", None):
            try:
                ref_page = get_related_object(self.config, "internal_link")
                link = ref_page.get_absolute_url()
            except (
                KeyError,
                TypeError,
                ValueError,
                AttributeError,
                models.ObjectDoesNotExist,
            ):
                self.internal_link = None
                return ""

            # simulate the call to the unauthorized CMSPlugin.page property
            cms_page = self.placeholder.page if self.placeholder_id else None

            # first, we check if the placeholder the plugin is attached to
            # has a page. Thus the check "is not None":
            if cms_page is not None:
                if getattr(cms_page, "node", None):
                    cms_page_site_id = getattr(cms_page.node, "site_id", None)
                else:
                    cms_page_site_id = getattr(cms_page, "site_id", None)
            # a plugin might not be attached to a page and thus has no site
            # associated with it. This also applies to plugins inside
            # static placeholders
            else:
                cms_page_site_id = None

            # now we do the same for the reference page the plugin links to
            # in order to compare them later
            if getattr(ref_page, "node", None) and cms_page is not None:
                ref_page_site_id = ref_page.node.site_id
            elif getattr(ref_page, "site_id", None) and cms_page is not None:
                ref_page_site_id = ref_page.site_id
            # if no external reference is found the plugin links to the
            # current page
            else:
                ref_page_site_id = Site.objects.get_current().pk

            if ref_page_site_id != cms_page_site_id:
                ref_site = Site.objects._get_site_by_id(ref_page_site_id).domain
                link = f"//{ref_site}{link}"

        elif getattr(self, "file_link", None):
            link = getattr(get_related_object(self.config, "file_link"), "url", "")

        elif getattr(self, "external_link", None):
            link = self.external_link

        elif getattr(self, "phone", None):
            link = "tel:{}".format(self.phone.replace(" ", ""))

        elif getattr(self, "mailto", None):
            link = f"mailto:{self.mailto}"

        else:
            link = ""

        if (
            not getattr(self, "phone", None) and not getattr(self, "mailto", None)
        ) and getattr(self, "anchor", None):
            link += f"#{self.anchor}"

        return link


class Link(GetLinkMixin, FrontendUIItem):
    """
    Components > "Button" Plugin
    https://getbootstrap.com/docs/5.0/components/buttons/
    """

    class Meta:
        proxy = True
        verbose_name = _("Link")

    def get_short_description(self):
        if self.config.get("name", "") and self.get_link():
            return f"{self.name} ({self.get_link()})"
        return self.config.get("name", "") or self.get_link() or _("<link is missing>")



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from djangocms_frontend import settings

from ...cms_plugins import CMSUIPlugin
from ...common.attributes import AttributesMixin
from ...common.spacing import SpacingMixin
from .. import utilities
from . import forms, models

mixin_factory = settings.get_renderer(utilities)


@plugin_pool.register_plugin
class SpacingPlugin(mixin_factory("Spacing"), AttributesMixin, CMSUIPlugin):
    """
    Components > "Card" Plugin
    https://getbootstrap.com/docs/5.0/components/card/
    """

    name = _("Spacing")
    module = _("Frontend")
    model = models.Spacing
    form = forms.SpacingForm

    change_form_template = "djangocms_frontend/admin/spacing.html"
    allow_children = True

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "space_property",
                    "space_sides",
                    "space_size",
                    "space_device",
                )
            },
        ),
    ]


@plugin_pool.register_plugin
class EditorNotePlugin(mixin_factory("EditorNote"), CMSUIPlugin):
    """Room for notes for editor only visible in edit mode"""

    name = _("Editor note")
    module = _("Frontend")
    allow_children = True
    change_form_template = "djangocms_frontend/admin/no_form.html"


@plugin_pool.register_plugin
class HeadingPlugin(
    mixin_factory("Heading"), AttributesMixin, SpacingMixin, CMSUIPlugin
):
    """Room for notes for editor only visible in edit mode"""

    name = _("Heading")
    module = _("Frontend")
    model = models.Heading
    form = forms.HeadingForm

    render_template = "djangocms_frontend/heading.html"
    allow_children = True

    fieldsets = [
        (
            None,
            {
                "fields": (
                    ("heading_level", "heading_id"),
                    "heading",
                    "heading_alignment",
                    "heading_context",
                )
            },
        ),
    ]

    def render(self, context, instance, placeholder):
        if not hasattr(context["request"], "TOC"):
            context["request"].TOC = []
        heading_id = getattr(instance, "heading_id", "")
        if heading_id:
            context["request"].TOC.append(
                (
                    heading_id,
                    getattr(instance, "heading", ""),
                    getattr(instance, "heading_level", "h2"),
                )
            )
        context["instance"] = instance
        return super().render(context, instance, placeholder)


def create_tree(request_toc):
    def process_level():
        nonlocal i

        previous_level = None
        toc_tree = []
        while i < len(request_toc):
            if previous_level is None or previous_level == request_toc[i][2]:
                toc_tree.append((request_toc[i][0], request_toc[i][1]))
                previous_level = request_toc[i][2]
                i += 1
            elif previous_level < request_toc[i][2]:
                toc_tree.append((None, process_level()))
            elif previous_level > request_toc[i][2]:
                break
        return toc_tree

    i = 0
    return process_level()


@plugin_pool.register_plugin
class TOCPlugin(mixin_factory("TOC"), AttributesMixin, CMSUIPlugin):
    name = _("Table of contents")
    module = _("Frontend")

    model = models.TableOfContents
    form = forms.TableOfContentsForm

    render_template = "djangocms_frontend/toc.html"
    change_form_template = "djangocms_frontend/admin/no_form.html"

    fieldsets = settings.EMPTY_FIELDSET

    def render(self, context, instance, placeholder):
        if hasattr(context["request"], "TOC"):
            toc_tree = create_tree(context["request"].TOC)
            context["toc"] = toc_tree
        else:
            context["toc"] = []
        context["template"] = self.render_template
        context["instance"] = instance
        return super().render(context, instance, placeholder)



from django.utils.translation import gettext_lazy as _

from djangocms_frontend.models import FrontendUIItem

# TODO leaving this comment for now
# data-bs-toggle="collapse" data-bs-target="#collapseExample"
# aria-expanded="false" aria-controls="collapseExample">
# data-bs-target can also be classes
# data-bs-parent links to the wrapper collapse
# <div class="collapse" id="collapseExample">


class Collapse(FrontendUIItem):
    """
    Component > "Collapse" Plugin
    https://getbootstrap.com/docs/5.0/components/collapse/
    """

    class Meta:
        proxy = True
        verbose_name = _("Collapse")

    def get_short_description(self):
        return f"(collapse-{str(self.pk)})"


class CollapseTrigger(FrontendUIItem):
    """
    Component > "Collapse Trigger" Plugin
    https://getbootstrap.com/docs/5.0/components/collapse/
    """

    class Meta:
        proxy = True
        verbose_name = _("Collapse trigger")

    def get_short_description(self):
        return f"({self.trigger_identifier})"


class CollapseContainer(FrontendUIItem):
    """
    Component > "Collapse Container" Plugin
    https://getbootstrap.com/docs/5.0/components/collapse/
    """

    class Meta:
        proxy = True
        verbose_name = _("Collapse container")

    def get_short_description(self):
        return f"({self.container_identifier})"



class ImageRenderMixin:
    def render(self, context, instance, placeholder):
        # assign link to a context variable to be performant
        context["picture_link"] = instance.get_link()
        context["picture_size"] = instance.get_size(
            width=context.get("width", 0),
            height=context.get("height", 0),
        )
        context["img_srcset_data"] = instance.img_srcset_data
        if instance.alignment:
            instance.add_classes(f"align-{instance.alignment}")
        if instance.picture_fluid:
            instance.add_classes("img-fluid")
        if instance.picture_rounded:
            instance.add_classes("rounded")
        if instance.picture_thumbnail:
            instance.add_classes("img-thumbnail")
        if instance.parent and instance.parent.plugin_type == "CardPlugin":
            instance.add_classes(
                "card-img-top" if instance.position == 0 else "card-img-bottom"
            )
        elif instance.parent and instance.parent.plugin_type == "FigurePlugin":
            instance.add_classes("figure-img")
        return super().render(context, instance, placeholder)



from django.test import TestCase

from djangocms_frontend.contrib.icon.models import Icon

icon_config = {
    "icon": {
        "libraryId": "zondicons",
        "libraryName": "zondicons",
        "iconHtml": '<i class="zi zi-airplane"></i>',
        "iconMarkup": "&lt;i class=&quot;zi zi-airplane&quot;&gt;&lt;/i&gt;",
        "iconClass": "zi zi-airplane",
        "iconText": "",
        "library": "zondicons",
    },
    "icon_size": "400%",
    "icon_foreground": "primary",
    "icon_rounded": True,
    "attributes": {"color": "black"},
    "background_context": "primary",
    "background_opacity": "50",
    "background_shadow": "reg",
    "responsive_visibility": None,
}


class IconModelTestCase(TestCase):
    def test_icon_instance(self):
        instance = Icon.objects.create(config=icon_config)
        self.assertEqual(str(instance), "Icon (1)")
        self.assertEqual(instance.get_short_description(), "zi zi-airplane")



from django.utils.translation import gettext_lazy as _

from djangocms_frontend.models import FrontendUIItem


class Accordion(FrontendUIItem):
    """
    Components > "Accordion" Plugin
    https://getbootstrap.com/docs/5.0/components/accordion/
    """

    class Meta:
        proxy = True
        verbose_name = _("Accordion")

    def get_children_count(self):
        return self.cmsplugin_set.count()

    def get_short_description(self):
        return _("({} entries)").format(self.get_children_count())


class AccordionItem(FrontendUIItem):
    """
    Components > "Accordion" Plugin
    https://getbootstrap.com/docs/5.0/components/accordion/
    """

    _("AccordionItem")

    class Meta:
        proxy = True
        verbose_name = _("Accordion item")

    def get_short_description(self):
        return getattr(self, "accordion_item_header", "-")



# Generated by Django 3.2.12 on 2022-03-16 17:18

from django.db import migrations


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("djangocms_frontend", "0001_initial"),
        ("link", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="NavBrand",
            fields=[],
            options={
                "verbose_name": "Brand",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("djangocms_frontend.frontenduiitem",),
        ),
        migrations.CreateModel(
            name="NavContainer",
            fields=[],
            options={
                "verbose_name": "Navigation container",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("djangocms_frontend.frontenduiitem",),
        ),
        migrations.CreateModel(
            name="Navigation",
            fields=[],
            options={
                "verbose_name": "Navigation",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("djangocms_frontend.frontenduiitem",),
        ),
        migrations.CreateModel(
            name="NavLink",
            fields=[],
            options={
                "verbose_name": "Navigation Link",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("link.link",),
        ),
        migrations.CreateModel(
            name="PageTree",
            fields=[],
            options={
                "verbose_name": "Page tree",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("djangocms_frontend.frontenduiitem",),
        ),
    ]



from django.test import TestCase

from djangocms_frontend.contrib.media.models import Media, MediaBody


class MediaModelTestCase(TestCase):
    def test_media_instance(self):
        instance = Media.objects.create()
        self.assertEqual(str(instance), "Media (1)")
        self.assertEqual(instance.get_short_description(), "")

    def test_media_body_instance(self):
        instance = MediaBody.objects.create()
        self.assertEqual(str(instance), "MediaBody (1)")
        self.assertEqual(instance.get_short_description(), "")
